/*->c.chain */

#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "transform.h"

#include "constants.h"
#include "str.h"
#include "etc.h"

#include "file.h"

#include "object.h"

#include "chain.h"


/* this is a funny function */
/* if after exists, the object is inserted after it in the chain */
/* otherwise the object is made the first child of parent        */

os_error * insafter(int object,int after,int parent,int coff,filestr * file)
{
 chainobject * link;
 char        * p;
 int           next;

 if(after)
 {
  link=(chainobject*)fob_find(after,file);
  next=link->chain.next;
  link->chain.next=object;

  if(next)
  {
   link=(chainobject*)fob_find(next,file);
   link->chain.prev=object;
  }

  link=(chainobject*)fob_find(object,file);
  link->chain.next=next;
  link->chain.prev=after;
 }
 else
 {
  if(coff>0)
  {
   p=(char*)fob_find(parent,file);
   p+=coff+2*sizeof(int);
   next=*((int*)p);
   *((int*)p)=object;
  }
  else
  if(coff==0)
  {
   link=(chainobject*)fob_find(parent,file);
   next=link->chain.child;
   link->chain.child=object;
  }
  else
  {
   next=0;
  }

  if(next)
  {
   link=(chainobject*)fob_find(next,file);
   link->chain.prev=object;
   link->chain.parent=0;
  }

  link=(chainobject*)fob_find(object,file);
  link->chain.parent=parent;
  link->chain.next=next;
  link->chain.prev=0;
 }
 return(NULL);
}



int endofchain(int first,filestr * file)
{
 chainobject * link;

 if(first)
 {
  while(1)
  {
   link=(chainobject*)fob_find(first,file);
   if(link->chain.next) first=link->chain.next;
   else                 break;
  }
 }
 return(first);
}



int startofchain(int first,filestr * file)
{
 chainobject * link;

 if(first)
 {
  while(1)
  {
   link=(chainobject*)fob_find(first,file);
   if(link->chain.prev) first=link->chain.prev;
   else                 break;
  }
 }
 return(first);
}


int parentof(chainobject * link,filestr * file)
{
 while(link->chain.prev) link=(chainobject*)fob_find(link->chain.prev,file);
 return(link->chain.parent);
}


int childobject(chainobject * link,filestr * file)
{
 objhdrstr * obj;
 int         parent;

 parent=parentof(link,file);
 if(parent)
 {
  obj=(objhdrstr*)fob_find(parent,file);
  return(obj->type!=OPAGE);
 }

 return(0);
}


int pageof(chainobject * link,filestr * file)
{
 while(link->type!=OPAGE)
 {
  while(link->chain.prev) link=(chainobject*)fob_find(link->chain.prev,file);
  if(link->chain.parent)  link=(chainobject*)fob_find(link->chain.parent,file);
  else                    return(0);
 }
 return(link->key);
}


int pagesectionof(chainobject * link,filestr * file,int * page,int * section)
{
 while(link->type!=OPAGE)
 {
  while(link->chain.prev) link=(chainobject*)fob_find(link->chain.prev,file);
  if(link->chain.parent)  link=(chainobject*)fob_find(link->chain.parent,file);
  else                    return(0);

/*  if(!link) return(0); */
 }
 *page=link->key;
 *section=((pagestr*)link)->section;
 return(1);
}


int sectionofpage(int page,filestr * file)
{
 return(((pagestr*)fob_find(page,file))->section);
}




os_error * unlinkobject(int object,int coff,filestr * file)
{
 chainobject * link;
 int           next;
 int           prev;
 int           parent;
 char        * p;

 link=(chainobject*)fob_find(object,file);

 next=link->chain.next;
 prev=link->chain.prev;
 parent=link->chain.parent;

 link->chain.next=link->chain.prev=link->chain.parent=0;

 if(parent)
 {
  if(coff)
  {
   p=(char*)fob_find(parent,file);
   p+=coff+2*sizeof(int);
   *((int*)p)=next;
  }
  else
  {
   link=(chainobject*)fob_find(parent,file);
   link->chain.child=next;
  }
 }

 if(prev)
 {
  link=(chainobject*)fob_find(prev,file);
  link->chain.next=next;
 }

 if(next)
 {
  link=(chainobject*)fob_find(next,file);
  link->chain.prev=prev;
  link->chain.parent=parent;
 }

 return(NULL);
}




static int xsobject;
static int xsindex;

/* as soon as you see a non-master object ack the message */

static os_error * searchfn(objhdrstr * obj,filestr * file,int * message)
{
 xsindex++;
 if(obj->key==xsobject) *message=0;
 return(NULL);
 USE(file);
}



int objectinobjectchain(int object,int child,filestr * file)
{
 os_error    * err;
 int           message;

 message=OM_FUNCTION;
 xsobject=object;

 err=objectchainmessage(child,file,&message,(void *)searchfn);
 if(err) return(0);

 return((message==0));
}


/* used to see if an object is linked into the current page */

int objectinobject(int object,int parent,filestr * file)
{
 chainobject * link;
 link=(chainobject*)fob_find(parent,file);
 return(objectinobjectchain(object,link->chain.child,file));
}



/* allow for child frames of embeded frames */

int embededobject(int object,filestr * file)
{
 chainobject * link;
 int           parent;

 link=(chainobject*)fob_find(object,file);
 if(link->chain.prev==NULL && link->chain.next==NULL)
 {
  parent=link->chain.parent;

  if(parent)
  {
   link=(chainobject*)fob_find(parent,file);
   if(link->chain.child==object) return(0);

   if(link->type==OPAGE && ((pagestr*)link)->gchain.child==object) return(0);

  }

  return(1);
 }

 return(0);
}



int objectchainindex(int object,int child,filestr * file)
{
 os_error    * err;
 int           message;

 message=OM_FUNCTION;
 xsobject=object;
 xsindex=0;

 err=objectchainmessage(child,file,&message,(void *)searchfn);
 if(err) return(0);

 return(message==0?xsindex:-1);
}



int objectindex(int object,int parent,filestr * file)
{
 chainobject * link;

 link=(chainobject*)fob_find(parent,file);

 return(objectchainindex(object,link->chain.child,file));
}



static int targetindex;


static os_error * indexfn(objhdrstr * obj,filestr * file,int * message)
{
 xsindex++;
 if(targetindex==xsindex)
 {
  *message=0;
  xsobject=obj->key;
 }
 return(NULL);
 USE(file);
}




int objectfindchainindex(int index,int child,filestr * file)
{
 os_error    * err;
 int           message;

 message=OM_FUNCTION;
 xsobject=0;
 xsindex=0;
 targetindex=index;

 err=objectchainmessage(child,file,&message,(void *)indexfn);
 if(err) return(0);

 return(xsobject);
}




int objectfindindex(int index,int parent,filestr * file)
{
 chainobject * link;

 link=(chainobject*)fob_find(parent,file);

 return(objectfindchainindex(index,link->chain.child,file));
}


#ifdef NEVER

os_error * dereferenceobject(int object,int * ref,filestr * file)
{
 chainobject * link;
 int           ni;
 int           np;
 int           parent;
 sectionstr  * sectionp;


 if(object==0) *ref=0;
 else
 {
  ni=np=0;
  parent=object;

  while(1)
  {
   link=fob_find(object,file);
 
   if(object!=parent && link->chain.child)
   {
    ni+=countchildobjects(link,file);
   }

   if(link->type==OPAGE)
   {
    np=((pagestr*)link)->data.number;
    sectionp=fob_find(((pagestr*)link)->section,file);
    np=np-sectionp->data.pagenumber.number+1;
    break;
   }
   if(link->chain.prev)
   {
    ni++;
    object=link->chain.prev;
   }
   else
   {
    if(link->chain.parent)
    {
     ni++;
     parent=object=link->chain.parent;
    }
    else break;
   }
  }

  if(np || ni) *ref=(0x80000000|(np<<16)|ni);
  else         *ref=object;
 }

 return(NULL);
}




os_error * referenceobject(int ref,int * object,filestr * file)
{
 int ni;
 int np;


 if(ref & 0x80000000)
 {
  ni=ref & 0xFFFF;
  np=(ref>>16) & 0x7FFF;






 }
 else *object=ref;

 return(NULL);
}


#endif





os_error * pagenext(filestr * file,int page,int * after,int * parent)
{
 pagestr * pagep;

 pagep=(pagestr*)fob_find(page,file);
 if(pagep->chain.child)
 {
  *parent=0;
  *after=endofchain(pagep->chain.child,file);
 }
 else
 {
  *parent=page;
  *after=0;
 }

 return(NULL);
}




os_error * changesize(chainobject * link,filestr * file)
{
 os_error * err;
 int        message;
 int        parent;

 err=NULL;

 message=OM_CHANGESIZE;
 parent=parentof(link,file);
 if(parent)
 {
  err=objectmessage((objhdrstr*)fob_find(parent,file),file,&message,NULL);
 }

 return(err);
}





static int xprotot;
static int xprototarget;

static os_error * pfn(sobjectstr * sob,filestr * file,int * message)
{
 int          master;
 objectbitstr bits;

 if(sob->type==xprotot)
 {
  xprototarget=sob->key;
  objectmaster(xprototarget,file,&master,&bits);
  if(!master) *message=0;
 }

 return(NULL);

 USE(file);
}



int xprototype(int type)
{
 int message;

 if(deffile)
 {
  message=OM_FUNCTION;
  xprotot=type;
  broadcastmessageb(deffile,DOC,&message,(void*)pfn);
  if(!message) return(xprototarget);
 }

 return(NULL);
}


