/*->c.bms */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <ctype.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "alloc.h"
#include "font.h"
#include "transform.h"

#include "etc.h"
#include "err.h"
#include "crc.h"
#include "xext.h"
#include "bits.h"
#include "xmath.h"
#include "fsx.h"
#include "bf.h"


#include "str.h"
#include "constants.h"

#include "reslink.h"
#include "xredraw.h"
#include "object.h"

#include "xcolour.h"

#include "file.h"
#include "view.h"
#include "undo.h"
#include "pref.h"


#include "txstr.h"
#include "act.h"
#include "vec.h"
#include "tx.h"
#include "find.h"
#include "wcx.h"


#include "bms.h"




#define BMNAMELEN 20



typedef struct bmstr
{
 char          name[BMNAMELEN];
 int           n;
 int           offset;
 int           source;
 int           text;
 int           u;
 int           flags;
 int           spare;

} bmstr;




typedef struct bmbasestr
{
 int           key;
 int           type;

 int           spare[4];

 int           n;
 int           sequence;
 bmstr         bm[1];

} bmbasestr;







static os_error * bmseek(filestr * file,char * name,int * handle)
{
 bmbasestr * bmb;
 int         i;

 *handle=0;

 if(file)
 {
  if(file->hdr.bms)
  { 
   bmb=(bmbasestr*)fob_find(file->hdr.bms,file);

   for(i=0;i<bmb->n;i++)
   {
    if(!strcmp(name,bmb->bm[i].name))
    {
     *handle=bmb->bm[i].n;
     break;
    }
   }
  }
  return(NULL);
 }
 else return(geterror(BM01));
}



static int bmcmpfn(const void * p1,const void * p2)
{
 return(((bmstr*)p1)->n-((bmstr*)p2)->n);
}


static os_error * bmlocate(filestr * file,int handle,
                                               int * index,bmbasestr ** bmbp)
{
 bmbasestr * bmb;
 bmstr       target;
 bmstr     * bp;

 *index=-1;

 if(file)
 {
  if(file->hdr.bms)
  { 
   *bmbp=bmb=(bmbasestr*)fob_find(file->hdr.bms,file);

   target.n=handle;

   bp=(bmstr*)bsearch(&target,bmb->bm,bmb->n,sizeof(bmstr),bmcmpfn);
   if(bp) *index=bp-bmb->bm;           
  }
  return(NULL);
 }
 else return(geterror(BM01));
}




static os_error * bmboot(filestr * file)
{
 os_error  * err;
 bmbasestr * bmb;

 err=NULL;

 if(file)
 {
  if(!file->hdr.bms)
  {
   err=fob_create(sizeof(bmbasestr),file,OBMS,(fobjptr)&bmb);
   if(!err)
   {
    file->hdr.bms=bmb->key;
    bmb->sequence=1;
   }
  }
 }
 else err=geterror(BM01);

 return(err);
}


static os_error * bmadd(filestr * file,char * name,int * handle)
{
 os_error  * err;
 bmbasestr * bmb;
 int         n;

 *handle=0;

 err=bmboot(file);
 if(!err)
 {
  err=bmseek(file,name,handle);
  if(!err && !(*handle))
  {
   if(file->hdr.bms)
   {
    bmb=(bmbasestr*)fob_find(file->hdr.bms,file);
    n=bmb->n;

    err=fob_change(file->hdr.bms,file,sizeof(bmbasestr)+n*sizeof(bmstr),
                                                             (fobjptr)&bmb);
    if(!err)
    {
     memset(bmb->bm+n,0,sizeof(bmstr));
     *handle=bmb->bm[n].n=bmb->sequence++;
     xzstrncpy(bmb->bm[n].name,name,BMNAMELEN);
     bmb->n++;
    }
   }
  }
 }

 return(err);
}


static os_error * bmrem(filestr * file,int handle)
{
 os_error  * err;
 bmbasestr * bmb;
 int         index;
 int         n;

 err=NULL;

 if(file)
 {
  if(file->hdr.bms && handle)
  {
   bmlocate(file,handle,&index,&bmb);
   if(index>=0)
   {
    memmove(bmb->bm+index,bmb->bm+index+1,(bmb->n-index)*sizeof(bmstr));
    n=bmb->n--;
    err=fob_change(file->hdr.bms,file,sizeof(bmbasestr)+(n-1)*sizeof(bmstr),
                                                             (fobjptr)&bmb);
   }
  }
 }
 else err=geterror(BM01);

 return(err);
}






static int starget;
static int stargetu;

static os_error * sfn(textboxstr * textp,filestr * file,int * message)
{
 os_error   * err;
 texsrc     * texsp;

 err=NULL;

 if(textp->type==OTEXTBOX)
 {
  texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));

  if(((stargetu==1 && textp->data.header) && texsp[1].source==starget) ||
     ((stargetu==2 && textp->data.footer) && texsp[2].source==starget) ||
     texsp->source==starget)
  {
   *message=0;
   starget=textp->key;
  }
 }

 return(err);

 USE(file);
}



static os_error * bmsetsouce(bmstr * bm,filestr * file)
{
 int message;

 starget=bm->source;
 stargetu=bm->u;
 message=OM_FUNCTION;
 broadcastmessage(file,&message,(void*)sfn);
 if(!message) bm->text=starget;

 return(NULL);
}



static os_error * bmsetvector(vector * s,filestr * file,
                                                   bmbasestr * bmb,int index)
{
 os_error * err;
 bmstr      bm;

 bm=bmb->bm[index];

 if(!bm.text && bm.source) bmsetsouce(&bm,file);


 if(!bm.text || !bm.source || bm.offset<0) err=geterror(BM00);
 else
 {
  if(act.valid && bm.source==act.textsrc.source && file==act.file)
  {
   *s=act;
  }
  else
  {
   setvectorfrombox(s,bm.u,bm.text,file);
  }

  err=movevector(s,bm.offset);
 }

 return(err);
}



os_error * bmsourcechange(textsrcstr * textsrc,filestr * file,int delta)
{
 bmbasestr * bmb;
 int         n;
 int         i;

 if(delta && file->hdr.bms)
 {
  bmb=(bmbasestr*)fob_find(file->hdr.bms,file);
  n=bmb->n;
  if(n)
  {
   for(i=0;i<n;i++)
   {
    if(bmb->bm[i].source==textsrc->source)
    {
     if(delta>0)
     {
      if(bmb->bm[i].offset>=textsrc->offset) bmb->bm[i].offset+=delta;
     }
     else
     {
      if(bmb->bm[i].offset>(textsrc->offset-delta)) bmb->bm[i].offset+=delta;
      else
      if(bmb->bm[i].offset>textsrc->offset) bmb->bm[i].offset=textsrc->offset;
     }
    }
   }
  }
 }

 return(NULL);
}


os_error * bmsourcedel(int source,filestr * file)
{
 bmbasestr * bmb;
 int         n;
 int         i;

 if(file->hdr.bms && source)
 {
  bmb=(bmbasestr*)fob_find(file->hdr.bms,file);
  n=bmb->n;
  if(n)
  {
   for(i=0;i<n;i++)
   {
    if(bmb->bm[i].source==source)
    {
     bmrem(file,bmb->bm[i].n);    /* not very efficient */
    }
   }
  }
 }
 return(NULL);
}




os_error * bmtextdel(int text,filestr * file)
{
 bmbasestr * bmb;
 int         n;
 int         i;

 if(file->hdr.bms && text)
 {
  bmb=(bmbasestr*)fob_find(file->hdr.bms,file);
  n=bmb->n;
  if(n)
  {
   for(i=0;i<n;i++)
   {
    if(bmb->bm[i].text==text)
    {
     bmb->bm[i].text=0;
    }
   }
  }
 }
 return(NULL);
}



/* set bm to vector */

static void asignbmvec(vector * v,bmbasestr * bmb,int index)
{
 bmb->bm[index].offset=v->textsrc.offset;
 bmb->bm[index].source=v->textsrc.source;
 bmb->bm[index].text=v->textsrc.text;
 bmb->bm[index].u=v->textsrc.u;
}


/* bm1 becomes bm2 */

static void asignbmbm(bmbasestr * bmb,int index1,int index2)
{
 bmb->bm[index1].source=bmb->bm[index2].source;
 bmb->bm[index1].offset=bmb->bm[index2].offset;
 bmb->bm[index1].text=bmb->bm[index2].text;
 bmb->bm[index1].u=bmb->bm[index2].u;
}


/*****************************************************************************/


/* int  bmcreate(string & name) Create bookmark with given name,return handle*/

int  bmcreate(int fp)
{
 char temp[BMNAMELEN];
 int  handle;

 xzstrncpy(temp,stringptr(stack[fp]),BMNAMELEN);
 zreport(bmadd(menufile,temp,&handle));

 return(handle);
}


/* void bmdelete(int bmhandle);              Delete a bookmark */


void bmdelete(int fp)
{
 zreport(bmrem(menufile,stack[fp]));
}


/* int bmfind(string & name);               Map bookmark name to handle */

int  bmfind(int fp)
{
 char temp[BMNAMELEN];
 int  handle;

 xzstrncpy(temp,stringptr(stack[fp]),BMNAMELEN);

 zreport(bmseek(menufile,temp,&handle));

 return(handle);
}


/* void setbmtocaret(int bmhandle);          Set book mark to caret posn */

void setbmtocaret(int fp)
{
 int         handle;
 bmbasestr * bmb;
 int         index;

 handle=stack[fp];

 if(menufile==act.file && act.valid)
 {
  bmlocate(menufile,handle,&index,&bmb);

  if(index>=0)
  {
   asignbmvec(&act,bmb,index);
  }
 }
}


static os_error * setcarettobms(int fp,int act)
{
 os_error  * err;
 int         handle;
 vector      s;
 bmbasestr * bmb;
 int         index;

 handle=stack[fp];

 err=bmlocate(menufile,handle,&index,&bmb);
 if(!err && index>=0)
 {
  err=bmsetvector(&s,menufile,bmb,index);
  if(!err)
  {
   err=txlock(&s.textsrc,s.file);
   if(!err)
   {
    vsetxy(&s); 
    screenvector(&s,menuview);
    txunlock(&s.textsrc,s.file);

    if(act) makeactive(&s);
   }
  }
 }

 return(err);
}



/* void setcarettobm(int bmhandle);          Set caret posn to bookmark */

void setcarettobm(int fp)
{
 zreport(setcarettobms(fp,1));
}


/* void bmview(int bmhandle);            Ensure book mark posn is visible */

void bmview(int fp)
{
 zreport(setcarettobms(fp,0));
}




/* void setzonetobm(int bmhandle); Set marked zone between bookmark and caret*/

void setzonetobm(int fp)
{
 os_error  * err;
 int         handle;
 vector      s1;
 vector      s2;
 bmbasestr * bmb;
 int         index;
 int         rzone;

 err=NULL;

 if(act.valid && act.file==menufile)
 {
  handle=stack[fp];

  bmlocate(menufile,handle,&index,&bmb);
  if(index>=0)
  {
   err=bmsetvector(&s2,menufile,bmb,index);
   if(!err && s2.textsrc.source==act.textsrc.source)
   {
    s1=act;

    nullzone(&s1,&s2,&rzone);

    if(zone)
    {
     err=undostart(menufile);
     if(!err)
     {
      unsaveact();

      if(rzone)
      {
       err=txlock(&s2.textsrc,s2.file);
       vsetxy(&s2);
       txunlock(&s2.textsrc,s2.file);

       s1=zon;
       zon=s2;
       refreshzone(&s1,&zon);
      }
      else
      {
       unselect();
      }

      undoend(menufile);
     }
    }
    else
    if(rzone)
    {
     setupzonehi(&s1,&s2);
    }
   }
  }
 }

 zreport(err);
}





static int bmchars(int fp,int prev)
{
 os_error  * err;
 int         handle;
 vector      s;
 bmbasestr * bmb;
 int         index;
 int         c;
	int         cv;

 err=NULL;

 handle=stack[fp];

 err=bmlocate(menufile,handle,&index,&bmb);
 if(!err && index>=0)
 {
  err=bmsetvector(&s,menufile,bmb,index);
  if(!err)
  {
   err=txlock(&s.textsrc,s.file);
   if(!err)
   {
    if(prev) c=prevch(&s,&cv);
    else     c=nextch(&s,&cv);
    txunlock(&s.textsrc,s.file);
    return(c);
   }
  }
 }

 zreport(err);
 return(-1);
}


/* int  bmchar(int bmhandle);                Return character at bookmark */

int bmchar(int fp)
{
 return(bmchars(fp,0));
}


int bmprevchar(int fp)
{
 return(bmchars(fp,1));
}




/* void setbmtobm(int bm1,int bm2);          Set posn of bm1 to that of bm2 */

void setbmtobm(int fp)
{
 os_error  * err;
 int         handle1;
 int         handle2;
 bmbasestr * bmb;
 int         index1=0;
 int         index2=0;

 handle1=stack[fp];
 handle2=stack[fp+1];

          err=bmlocate(menufile,handle1,&index1,&bmb);
 if(!err) err=bmlocate(menufile,handle2,&index2,&bmb);

 if(index1>=0 && index2>=0)
 {
  asignbmbm(bmb,index1,index2);
 }

 zreport(err);
}




static os_error * bmvectordown(vector * s,int * code)
{
 vsetxy(s);
 return(vectordown(s,s->x,1,code));
}


static os_error * bmvectorup(vector * s,int * code)
{
 vsetxy(s);
 return(vectorup(s,s->x,1,code));
}




/* int bmmove(int bhandle,int direction,int unit);             */
/* Move bookmark, forward/backward, by char,word,line           */


int bmmove(int fp)
{
 os_error  * err;
 int         handle;
 int         dir;
 int         unit;
 bmbasestr * bmb;
 int         index;
 int         code;
	int         codev;
 vector      s;

 err=NULL;

 handle=stack[fp];
 dir=stack[fp+1];
 unit=stack[fp+2];
 code=0;

 err=bmlocate(menufile,handle,&index,&bmb);
 if(!err && index>=0)
 {
  err=bmsetvector(&s,menufile,bmb,index);
  if(!err)
  {
   err=txlock(&s.textsrc,s.file);
   if(!err)
   {
    if(dir)
    {
     switch(unit)
     {
      case 0:
             code=nextch(&s,&codev);
             break;

      case 1:
             settoeow(&s);
             break;

      case 2:
             settoeol(&s,1);
             break;

      case 3:
             code=settoeop(&s);
             break;

      case 4:
             settoeot(&s);
             break;

      case 5:
             settosnw(&s);
             break;

      case 6:
             code=nextvline(&s);
             break;

      case 7:
             code=settosnp(&s);
             break;

      case 8:
             bmvectordown(&s,&code);
             break;
     }
    }
    else
    {
     switch(unit)
     {

      case 0:
             code=prevch(&s,&codev);
             break;

      case 1:
             settosow(&s);
             break;

      case 2:
             settosol(&s);
             break;

      case 3:
             code=settosop(&s);
             break;

      case 4:
             settosot(&s);
             break;

      case 5:
             settospw(&s);
             break;

      case 6:
             code=prevvline(&s);
             break;

      case 7:
             code=settospp(&s);
             break;

      case 8:
             bmvectorup(&s,&code);
             break;
     }
    }

    txunlock(&s.textsrc,s.file);

    bmb=(bmbasestr*)fob_find(s.file->hdr.bms,s.file);
    bmb->bm[index].offset=s.textsrc.offset;
    bmb->bm[index].text=s.textsrc.text;
   }
  }
 }

 zreport(err);
 return(code);
}



/* int  bmname(int n,string & name);                            */
/* Return name/handle of nth. bookmark, return zero for no more */


int bmname(int fp)
{
 int         n;
 int         s;
 int         handle;
 bmbasestr * bmb;
 char        temp[BMNAMELEN];

 n=stack[fp];
 s=stack[fp+1];

 handle=0;

 if(!menufile) zreport(geterror(BM01));

 if(menufile->hdr.bms)
 {
  bmb=(bmbasestr*)fob_find(menufile->hdr.bms,menufile);
  if(n>=0 && n<bmb->n)
  {
   handle=bmb->bm[n].n;
   xzstrncpy(temp,bmb->bm[n].name,sizeof(temp));
   assignstring(s,temp);
  }
 }

 return(handle);
}



/* void setbmtozone(int bm1,int bm2);        Set bookmarks to marked zone. */

void setbmtozone(int fp)
{
 int         handle1;
 int         handle2;
 bmbasestr * bmb;
 int         index1;
 int         index2;

 handle1=stack[fp];
 handle2=stack[fp+1];

 if(menufile==act.file && act.valid && zone)
 {
  bmlocate(menufile,handle1,&index1,&bmb);
  bmlocate(menufile,handle2,&index2,&bmb);

  if(index1>=0 && index2>=0)
  {
   asignbmvec(lozon,bmb,index1);
   asignbmvec(hizon,bmb,index2);
  }
 }
}



/* int  bmcmp(int bmhandle,int bmhandle);            */
/*  Compare posn of two bookmarks, return -1, 0, 1.  */


int bmcmp(int fp)
{
 os_error  * err;
 int         handle1;
 int         handle2;
 bmbasestr * bmb;
 int         index1=0;
 int         index2=0;
 int         temp;


 handle1=stack[fp];
 handle2=stack[fp+1];

          err=bmlocate(menufile,handle1,&index1,&bmb);
 if(!err) err=bmlocate(menufile,handle2,&index2,&bmb);

 if(!err && index1>=0 && index2>=0)
 {
  if(bmb->bm[index1].source==bmb->bm[index2].source)
  {
   temp=bmb->bm[index1].offset-bmb->bm[index2].offset;
   if(temp) return(SGN(temp));
   else     return(temp);
  }
  else return(-2);
 }

 zreport(err);

 return(-2);
}




/***************************************************************************/

static void getfindbits(findstr * find,int bits)
{
 memset(find,0,sizeof(findstr));

 find->xcase=bits & 0x1;
 find->whole=(bits>>1) & 0x1;
 find->all=(bits>>2) & 0x1;
 find->start=(bits>>3) & 0x1;
 find->wild=(bits>>4) & 0x1;
}


/* int bmsearch(int bm1,int bm2,string & s,int flags); */


int bmsearch(int fp)
{
 os_error  * err;
 findstr     find;
 int         found=0;
 int         soffset;
 int         eoffset;
 bmbasestr * bmb;
 vector      s;
 int         handle1;
 int         handle2;
 int         index1=0;
 int         index2=0;
 vector      e;
 sourcestr * sourcep;


 err=NULL;

 handle1=stack[fp];
 handle2=stack[fp+1];
 getfindbits(&find,stack[fp+3]);          /* zeros find structure */
 xzstrncpy(find.s,stringptr(stack[fp+2]),sizeof(find.s));

          err=bmlocate(menufile,handle1,&index1,&bmb);
 if(!err) err=bmlocate(menufile,handle2,&index2,&bmb);

 if(!err && index1>=0 && index2>=0)
 {
  err=bmsetvector(&s,menufile,bmb,index1);
  if(!err)
  {
   e=s;
   sourcep=(sourcestr*)fob_find(s.textsrc.source,s.file);
   e.textsrc.offset=sourcep->len;
 
   setupfind(&s,&e);

   err=search(&find,&found,&soffset,&eoffset);

   if(found && !err)
   {
    bmb=(bmbasestr*)fob_find(menufile->hdr.bms,menufile);

    bmb->bm[index1].offset=soffset;
    asignbmbm(bmb,index2,index1);
    bmb->bm[index2].offset=eoffset;
   }
  }
 }

 zreport(err);
 return(!found);
}



/* int  replace(string & s,int flags); */

int replace(int fp)
{
	os_error * err;
 findstr    find;

 getfindbits(&find,stack[fp+1]);    /* zeros find structure */

 if(act.valid && zone && act.file==menufile)
 {
  err=xzstrncpye(find.r,stringptr(stack[fp]),sizeof(find.r));
		if(!err)
		{
   saveinitfind3(lozon,lozon,hizon);

   err=freplace(&find);
		}

		zreport(err);

  return(0);
 }
 else return(1);
}




/* void bmstartscan(void);  prepare to scan all stories in document */

void bmstartscan(int fp)
{
 if(!menufile) zreport(geterror(BM01));
 firststory(menuview->branch,menufile);
 USE(fp);
}



/* int  bmnextstory(int bm);   move bm to start of next story
                               return non-zero when no more   */


int bmnextstory(int fp)
{
 int         handle;
 bmbasestr * bmb;
 int         index;
 vector      s;
 int         eof;

 handle=stack[fp];

 nextstory(&s,&eof);

 if(!eof)
 {
  zreport(bmlocate(menufile,handle,&index,&bmb));

  if(index>=0)
  {
   asignbmvec(&s,bmb,index);
			return(0);
  }
 }

	return(1);
}


int zoneorder(int fp)
{
 if(!zone || !act.valid || menufile!=act.file) return(0);
 else                                          return(zon.textsrc.offset-
                                                      act.textsrc.offset);
 USE(fp);
}



static int bmpagesect(int fp,int page)
{
 os_error   * err;
 int          handle;
 vector       s;
 bmbasestr  * bmb;
 int          index;
 pagestr    * pagep;
 sectionstr * sectionp;


 handle=stack[fp];

 err=bmlocate(menufile,handle,&index,&bmb);
 if(!err && index>=0)
 {
  err=bmsetvector(&s,menufile,bmb,index);
  if(!err)
  {
   if(page)
   {
    sectionp=(sectionstr*)fob_find(s.section,menufile);
    pagep=(pagestr*)fob_find(s.page,menufile);
    return(pagep->data.number-sectionp->data.pagenumber.number);
   }
   else
   {
    sectionp=(sectionstr*)fob_find(s.section,menufile);
    return(sectionp->data.number-prefp(menufile)->chapterstart);
   }
  }
 }

 zreport(err);
 return(-1);
}


int bmpage(int fp)
{
 return(bmpagesect(fp,1));
}

int bmsection(int fp)
{
 return(bmpagesect(fp,0));
}


/*************************************************************************/

#define BMHCHUNK 0x100

static char * bmhiddenlist;
static int    bmhiddenlen;


static os_error * bmhiddenadd(char * prefix)
{
 os_error * err;
 int        len;

 len=bmhiddenlen+strlen(prefix)+1;

 err=schunk((flex_ptr)&bmhiddenlist,len+1,BMHCHUNK);
 if(!err)
 {
  strcpy(bmhiddenlist+bmhiddenlen,prefix);
  *(bmhiddenlist+bmhiddenlen+1)=0;
  bmhiddenlen=len;
 }

 return(err);
}


static os_error * bmhiddenrem(int offset)
{
 int        len;

 len=strlen(bmhiddenlist+offset);
 memcpy(bmhiddenlist+offset,bmhiddenlist+offset+len+1,bmhiddenlen+1-offset);
 bmhiddenlen-=len;

 return(NULL);
}


static int bmhidden(char * prefix)
{
 char * p;

 p=bmhiddenlist;
 if(p)
 {
  while(*p)
  {
   if(strcmp(p,prefix)==0)
   {
    return(p-bmhiddenlist+1);
   }

   p+=strlen(p)+1;
  }
 }

 return(0);
}


//int bm_hidden(string &prefix)

int bm_hidden(int fp)
{
 return(bmhidden(stringptr(stack[fp])));
}


//void bm_hide(string &prefix)

void bm_hide(int fp)
{
 if(!bmhidden(stringptr(stack[fp])))
 {
  bmhiddenadd(stringptr(stack[fp]));
 }
}


//void bm_unhide(string &prefix)

void bm_unhide(int fp)
{
 int mrk;

 mrk=bmhidden(stringptr(stack[fp]));
 if(mrk)
 {
  bmhiddenrem(mrk-1);
 }
}



//void bm_toggle(string &prefix)

void bm_toggle(int fp)
{
 if(bm_hidden(fp)) bm_unhide(fp);
 else              bm_hide(fp);
}



// int bm_hiddenmark(string &name)

int bm_hiddenmark(int fp)
{
 char * p;
 int    len;

 p=bmhiddenlist;
 if(p)
 {
  while(*p)
  {
   len=strlen(p);

   if(strncmp(p,stringptr(stack[fp]),len)==0)
   {
    return(1);
   }

   p+=len+1;
  }
 }

 return(0);
}




/*************************************************************************/



os_error * bminit(void)
{



 return(NULL);
}

