/*->c.fmap */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  
#include <ctype.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "bbc.h"
#include "akbd.h"
#include "font.h"

#include "transform.h"
#include "poll.h"
#include "key.h"
#include "etc.h"
#include "err.h"
#include "hourglass.h"
#include "bits.h"
#include "fsx.h"
#include "bf.h"
#include "timex.h"
#include "crc.h"
#include "deb.h"


#include "str.h"
#include "constants.h"

#include "reslink.h"

#include "file.h"

#include "xprint.h"

#include "txstr.h"
#include "fox.h"
#include "fcache.h"
#include "style.h"

#include "fmapstr.h"
#include "fmap.h"



/*****************************************************************************/
                             /* Font Mapping */


/* idea is for each file, provide a font map                 */
/* hold all font names for file, usages, keys for font names */


int cmpfmapn(const void * s1,const void * s2)
{
 return(((fmaphdrstr*)s1)->key-((fmaphdrstr*)s2)->key);
}



os_error * fmapdup(filestr * file,int key)
{
 os_error       * err;
 fmapbasestr    * fbase;
 fmaphdrstr     * r;
 fmaphdrstr       fmaphdr;
 int              fmap;
 static int       ecount;


 fmap=file->hdr.fmap;
 err=fob_mod(fmap,file,(fobjptr)&fbase);
 if(!err)
 {
  fmaphdr.key=key;

  r=(fmaphdrstr*)bsearch(&fmaphdr,&fbase->fmaphdr[0],
                                   fbase->n,sizeof(fmaphdrstr),cmpfmapn);
  if(r)
  {
   r->count++;
  }
  else
  {
   if(!ecount)
   {
    err=geterror(STYLE04);
    ecount++;
   }
  }
 }
 return(err);
}




os_error * fmaprem(filestr * file,int key,int number)
{
 os_error       * err;
 fmapbasestr    * fbase;
 fmaphdrstr     * r;
 fmaphdrstr     * m;
 fmaphdrstr       fmaphdr;
 int              fmap;
 int              size;
 int              maxoffset;
 int              offset;
 int              n;
 int              i;
 static int       ecount;


 fmap=file->hdr.fmap;
 err=fob_mod(fmap,file,(fobjptr)&fbase);
 if(!err)
 {
  fmaphdr.key=key;

  r=(fmaphdrstr*)bsearch(&fmaphdr,&fbase->fmaphdr[0],
                                   fbase->n,sizeof(fmaphdrstr),cmpfmapn);
  if(r)
  {
   r->count-=number;
   if(r->count<=0)
   {
    size=strlen((char*)fbase->fmaphdr+r->offset)+1+sizeof(fmaphdrstr);
    offset=r->offset;
    m=&fbase->fmaphdr[fbase->n-1];
    maxoffset=m->offset+strlen((char*)fbase->fmaphdr+m->offset)+1;
    n=r-fbase->fmaphdr;

    for(i=n;i<(fbase->n-1);i++)
    {
     fbase->fmaphdr[i]=fbase->fmaphdr[i+1];
     fbase->fmaphdr[i].offset-=size;
    }
    for(i=0;i<n;i++)
    {
     fbase->fmaphdr[i].offset-=sizeof(fmaphdrstr);
    }

    memmove((char*)fbase->fmaphdr+(fbase->n-1)*sizeof(fmaphdrstr),
            (char*)fbase->fmaphdr+fbase->n*sizeof(fmaphdrstr),
             offset-fbase->n*sizeof(fmaphdrstr));

    offset-=sizeof(fmaphdrstr);
    memmove((char*)fbase->fmaphdr+offset,
            (char*)fbase->fmaphdr+offset+size,maxoffset-offset-size);

    maxoffset-=size;
    size=(int)sizeof(fmapbasestr)-(int)sizeof(fmaphdrstr)+maxoffset;
    err=fob_change(fmap,file,size,(fobjptr)&fbase);
    fbase->n--;
   }
  }
  else
  {
   if(!ecount)
   {
    err=geterror(STYLE04);
    ecount++;
   }
  }
 }
 return(err);
}



static int fmapcrc(char * s)
{
 int crc;

 crc=0;
 while(*s) crc=crcbyte32(crc,tolower(*s++));

 crc&=MAXFINT;
 if(crc<MINFINT) crc+=PRFINT;
 return(crc);
}


os_error * fmapadd(filestr * file,int ifn,int number,int * nkey)
{
 os_error      * err;
 fmapbasestr   * fbase;
 fmaphdrstr      fmaphdr;
 int             fmap;
 fmaphdrstr    * s;
 fmaphdrstr    * m;
 int             key;
 int             n;
 int             size;
 int             i;
 int             offset;
 int             maxoffset;
 int             minoffset;
 int             len;
 char          * name;



 fmap=file->hdr.fmap;

 err=fob_mod(fmap,file,(fobjptr)&fbase);
 if(!err)
 {
  name=fontidentifieri(ifn);

  len=strlen(name);
  key=fmapcrc(name);

  while(1)
  {
   fmaphdr.key=key;
   s=(fmaphdrstr*)bnearest(&fmaphdr,&fbase->fmaphdr[0],
                                 fbase->n,sizeof(fmaphdrstr),cmpfmapn);
   n=s-fbase->fmaphdr;

   if(n<fbase->n && !cmpfmapn(s,&fmaphdr))
   {                              /* equal numbers */
    if(cstrcmp((char*)fbase->fmaphdr+s->offset,name))
    {                             /* but different chains - a collision */
     key+=PRFINT;
     if(key>MAXFINT) key-=MAXFINT;
     if(key<MINFINT) key+=PRFINT;
    }
    else                            /* identical styles */
    {
     s->count+=number;
     break;
    }
   }
   else
   {
    if(fbase->n)
    {
     m=&fbase->fmaphdr[fbase->n-1];
     maxoffset=m->offset+strlen((char*)fbase->fmaphdr+m->offset)+1;
     size=maxoffset+sizeof(fmapbasestr)-sizeof(fmaphdr);
     size+=sizeof(fmaphdr)+len+1;
    }
    else
    {
     size=sizeof(fmapbasestr)+len+1;
     maxoffset=0;
    }

    err=fob_change(fmap,file,size,(fobjptr)&fbase);
    size=len+1;

    if(n<fbase->n)
    {
     offset=fbase->fmaphdr[n].offset;
     memmove(((char*)fbase->fmaphdr)+offset+sizeof(fmaphdr)+size,
            ((char*)fbase->fmaphdr)+offset,maxoffset-offset);
    }
    else
    {
     offset=maxoffset;
    }


    if(fbase->n)
    {
     minoffset=fbase->fmaphdr[0].offset;
     memmove(((char*)fbase->fmaphdr)+minoffset+sizeof(fmaphdr),
             ((char*)fbase->fmaphdr)+minoffset,offset-minoffset);
     offset+=sizeof(fmaphdr);

     for(i=fbase->n-1;i>=n;i--)
     {
      fbase->fmaphdr[i+1]=fbase->fmaphdr[i];
      fbase->fmaphdr[i+1].offset+=size+sizeof(fmaphdrstr);
     }
     for(i=0;i<n;i++)
     {
      fbase->fmaphdr[i].offset+=sizeof(fmaphdrstr);
     }
    }
    else
    {
     offset=sizeof(fmaphdrstr);
    }

    fbase->fmaphdr[n].offset=offset;
    fbase->fmaphdr[n].key=key;
    fbase->fmaphdr[n].count=number;

    fbase->fmaphdr[n].ifn=ifn;

    memmove(((char*)fbase->fmaphdr)+offset,fontidentifieri(ifn),size);
    fbase->n++;
    break;
   }
  }

  *nkey=key;
 }

 return(err);
}


char * fmapname(int i,filestr * file)
{
 fmapbasestr    * fbase;
 int              fmap;


 fmap=file->hdr.fmap;
 fbase=(fmapbasestr*)fob_find(fmap,file);
 return(((char*)fbase->fmaphdr)+fbase->fmaphdr[i].offset);
}


os_error * fmapremap(filestr * file,int * np,int * firstp,int * cn)
{
 os_error       * err;
 fmapbasestr    * fbase;
 int              fmap;
 int              i;
 int              first;
 int              n;
 int              changed;
 int              oldmap;
 int              newmap;


 first=n=changed=0;


 fmap=file->hdr.fmap;
 err=fob_mod(fmap,file,(fobjptr)&fbase);
 if(!err)
 {
  for(i=0;i<fbase->n;i++)
  {
   fbase->fmaphdr[i].ifn=findxuserfont(((char*)fbase->fmaphdr)+fbase->fmaphdr[i].offset);

   oldmap=fbase->fmaphdr[i].mapped;

   if(fbase->fmaphdr[i].ifn==-1)
   {
    if(!first) 
    {
     *firstp=i;
     first=1;
    }
    n++;
    newmap=1;
   }
   else
   {
    newmap=0;
   }

   fbase->fmaphdr[i].mapped=newmap;

   if(oldmap!=newmap) changed++;
  }
 }

 *np=n;
 *cn=changed;
 return(err);
}






typedef struct missfntstr
{
 char * name;
 int    claim;
} missfntstr;



os_error * fmapdeclarefonts(filestr * file,int * cn)
{
 os_error       * err;
 fmapbasestr    * fbase;
 int              changed;
 missfntstr       block;
 char             name[256];
 int              i;

 changed=0;

 err=fob_lock((fobjptr)&fbase,file->hdr.fmap,file);
 if(!err)
 {
  for(i=0;i<fbase->n;i++)
  {
   if(fbase->fmaphdr[i].ifn==-1)
   {
    xzstrncpy(name,((char*)fbase->fmaphdr)+fbase->fmaphdr[i].offset,sizeof(name));
    block.name=name;
    block.claim=0;

    event(EVENT_FONTMISSING,file,(int)&block);
    changed|=(block.claim!=0);
   }
  }
  fob_unlock((fobjptr)&fbase,file);
 }

 *cn=changed;
 return(err);
}





/* replace any -1 ifn's with a real font number */

#ifdef NEVER

os_error * fmapfix(filestr * file)
{
 os_error       * err;
 fmapbasestr    * fbase;
 int              fmap;
 int              i;
 int              font;


 fmap=file->hdr.fmap;
 err=fob_mod(fmap,file,(fobjptr)&fbase);
 if(!err)
 {
  font=defltfont();

  for(i=0;i<fbase->n;i++)
  {
   if(fbase->fmaphdr[i].ifn==-1) fbase->fmaphdr[i].ifn=font;
  }
 }
 return(err);
}

#endif




os_error * fmapabsstylebase(filestr * file,int fix,int * changesp)
{
 os_error        * err;
 fmapbasestr     * fbase;
 absstylebasestr * sbase;
 fmaphdrstr        fmaphdr;
 fmaphdrstr      * f1;
 int               fmap;
 int               i;
 int               font;
 int               xfont;
 int               changes;
 int               count;
 int               newkey;
 int               astyles;
 int               ifn;


 changes=0;


 fmap=file->hdr.fmap;
 astyles=file->hdr.astyles;

 err=fob_lock((fobjptr)&sbase,astyles,file);
 if(!err)
 {
  err=fob_lock((fobjptr)&fbase,fmap,file);
  if(!err)
  {
   font=defltfont();

   for(i=0;i<sbase->n;i++)
   {
    fmaphdr.key=sbase->absstyle[i].style.data.mfont;

    f1=(fmaphdrstr*)bsearch(&fmaphdr,fbase->fmaphdr,fbase->n,
                                          sizeof(fmaphdrstr),cmpfmapn);


    if(f1==NULL || f1->ifn==-1)
    {
     xfont=font;
     
     if(f1)
     {
      xfont=findreplacementfont(((char*)fbase->fmaphdr)+f1->offset);
      if(xfont==-1) xfont=font;
     }
     
     ifn=fontuserfromstyle2(xfont,sbase->absstyle[i].style.data.bold,
                                  sbase->absstyle[i].style.data.italic);

     if(fix)
     {
      count=sbase->absstyle[i].count;

      fmapadd(file,ifn,count,&newkey);
      fmaprem(file,sbase->absstyle[i].style.data.mfont,count);
      sbase->absstyle[i].style.data.mfont=newkey;
      sbase->absstyle[i].style.data.font=ifn;

      sbase->absstyle[i].changed=1;
      changes++;
     }
     else
     {
      sbase->absstyle[i].style.data.font=ifn;
     }
    }
    else
    {
     sbase->absstyle[i].style.data.font=f1->ifn;
    }
   }
   fob_unlock((fobjptr)&fbase,file);
  }
  fob_unlock((fobjptr)&sbase,file);
 }

 *changesp=changes;

 return(err);
}







os_error * fmapuserstylebase(filestr * file,int fix,int * changesp)
{
 os_error         * err;
 fmapbasestr      * fbase;
 userstylebasestr * sbase;
 fmaphdrstr         fmaphdr;
 fmaphdrstr       * f1;
 int                fmap;
 int                i;
 int                font;
 int                xfont;
 int                changes;
 int                count;
 int                newkey;
 int                ustyles;
 int                ifn;



 changes=0;


 fmap=file->hdr.fmap;
 ustyles=file->hdr.ustyles;
 err=fob_lock((fobjptr)&sbase,ustyles,file);
 if(!err)
 {
  err=fob_lock((fobjptr)&fbase,fmap,file);
  if(!err)
  {
   font=defltfont();

   for(i=0;i<sbase->n;i++)
   {
    fmaphdr.key=sbase->userstyle[i].style.data.mfont;

    f1=(fmaphdrstr*)bsearch(&fmaphdr,fbase->fmaphdr,fbase->n,
                                          sizeof(fmaphdrstr),cmpfmapn);

    if(f1==NULL || f1->ifn==-1)
    {
     xfont=font;
     
     if(f1)
     {
      xfont=findreplacementfont(((char*)fbase->fmaphdr)+f1->offset);
      if(xfont==-1) xfont=font;
     }
     
     
     ifn=fontuserfromstyle2(xfont,sbase->userstyle[i].style.data.bold,
                                  sbase->userstyle[i].style.data.italic);

     if(fix)
     {
      count=1;

      fmapadd(file,ifn,count,&newkey);
      fmaprem(file,sbase->userstyle[i].style.data.mfont,count);

      sbase->userstyle[i].style.data.mfont=newkey;
      sbase->userstyle[i].style.data.font=ifn;

      sbase->userstyle[i].changed=1;
      changes++;
     }
     else
     {
      sbase->userstyle[i].style.data.font=ifn;
     }
    }
    else
    {
     sbase->userstyle[i].style.data.font=f1->ifn;
    }
   }
   fob_unlock((fobjptr)&fbase,file);
  }
  fob_unlock((fobjptr)&sbase,file);
 }
 *changesp=changes;

 return(err);
}



/* used after missing fonts replaced, will only do something if fonts */
/* miss counted */

os_error * fmaptrim(filestr * file)
{
 os_error    * err;
 fmaphdrstr  * f1;
 fmapbasestr * fbase;
 int           n;
 int           fmap;
 int           nt;


 fmap=file->hdr.fmap;
 err=fob_lock((fobjptr)&fbase,fmap,file);
 if(!err)
 {
  for(n=0;n<fbase->n;n++)
  {
   f1=fbase->fmaphdr+n;

   if(f1->ifn==-1)
   {
    nt=fbase->n;
    fmaprem(file,f1->key,f1->count);
    if(fbase->n<nt)
    {
     n--;
    }
   }
  }

  fob_unlock((fobjptr)&fbase,file);
 }

 return(err);
}




os_error * newfmap(filestr * file)
{
 os_error      * err;
 int             fmapk=0; /* init to make compiler happy */
 fmapbasestr   * fmap;

 err=fob_create(sizeof(fmapbasestr),file,OFONTMAP,(fobjptr)&fmap);
 if(!err)
 {
  fmapk=fmap->key;
 }

 if(!err)
 {
  file->hdr.fmap=fmapk;
 }
 return(err);
}




/*
os_error * fmapdeclare(filestr * file)
{
 os_error       * err;
 fmapbasestr    * fbase;
 int              fmap;
 int              i;
	void           * cachehandle;

 err=NULL;

	cprintf("fmap decl in");

 fmap=file->hdr.fmap;
 fbase=(fmapbasestr*)fob_find(fmap,file);
 if(fbase)
 {
  for(i=0;i<fbase->n;i++)
  {
   err=printercachefont(((char*)fbase->fmaphdr)+fbase->fmaphdr[i].offset,0,&cachehandle);
			if(!err) err=printercachefontchar(32,255,cachehandle);
  }
 }

	cprintf("fmap decl ex");

 return(err); 
}
*/

