/*->c.fox */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  
#include <ctype.h>
#include <assert.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "flex.h"
#include "alloc.h"
#include "bbc.h"
#include "font.h"
#include "akbd.h"
#include "transform.h"

#include "osunits.h"
#include "osgraphics.h"


#include "wos.h"
#include "poll.h"
#include "key.h"
#include "etc.h"
#include "err.h"
#include "hourglass.h"
#include "xext.h"
#include "bits.h"
#include "dbhi.h"
#include "trans.h"
#include "deb.h"
#include "bf.h"

#include "str.h"
#include "constants.h"

#include "reslink.h"

#include "file.h"
#include "utils.h"


#include "txstr.h"
#include "tx.h"
#include "fcache.h"
#include "fx.h"
#include "style.h"
#include "fmap.h"
#include "dbchars.h"
#include "codes.h"
#include "txchoices.h"
#include "fxdraw.h"

#include "fox.h"



#ifdef _DEBUG

#define CHECKOSFONT(a) assert((a)>=0 && (a)<totosfonts);
 
#else

#define CHECKOSFONT(a)

#endif




typedef struct fontosnamestr
{
 int          foffset;    // offset to name
 int          fuoffset;   // offset to family name

 unsigned int unit:1;     // font family name
 unsigned int bold:1;
 unsigned int italic:1;
 unsigned int normal:1;
 unsigned int dataread:1;
 unsigned int weight:11;  // 16 bits total to here
 unsigned int charset:8;  // only need 8 for this really
 unsigned int pitch:1;
 unsigned int family:4;
 unsigned int ignore:1;
 unsigned int type:2; 

 int          ascent;
 int          descent;
 int          height;
 int          intleading;
 int          extleading;
 int          maxwidth;
 int          averagewidth;

 int          designsize;

 void       * data;
 void       * kdata;
 int          npairs;
 int          nchars;
 int        * glyphs;
 int          nglyph;  // number of pairs in glyphs

// unsigned int name:1;    this was used to split font names from id's

} fontosnamestr;



#define FOCHUNK (256*sizeof(fontosnamestr))
#define FONAMECHUNK 0x1000


static fontosnamestr * fontsos;
static int             totosfonts;

static void          * fontosnames;

extern os_error * stringheapfinit(void ** handle);
extern os_error * stringheapinit(void ** handle,int chunk);
extern os_error * stringheapadd(void ** handle,char * string,int * offset);
extern char     * stringheapstring(void * handle,int offset);
extern os_error * stringheapsize(void ** handle,int len,int * offset);




int           fontbits;

oshdc         fonthdcptr;





fontusernamestr * fontsuser;
int               totuserfonts;



static int fontignorefont(char * fontid);


/*****************************************************************************/



os_error * fontglyphs(int osfont,int ** glyphs,int * nglyph)
{
	os_error * err;

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
	{
 	*glyphs=fontsos[osfont].glyphs;

  *nglyph=fontsos[osfont].nglyph;
	}

	return(err);
}




int fontcharset(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].charset);
}


int fontweight(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].weight);
}


int fonttype(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].type);
}


int fontitalic(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].italic);
}


int fontpitch(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].pitch);
}


int fontfamily(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].family);
}


int fontaveragewidth(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].averagewidth);
}


int fontdesignsize(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].designsize);
}


int fontnpairs(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].npairs);
}



void * fontkdata(int i)
{
 CHECKOSFONT(i);

 return(fontsos[i].kdata);
}



int fontostotal(void)
{

 return(totosfonts);
}




typedef struct chardatastr
{
 int       c;
 int       w;
 int       h;
 font_info fi;

} chardatastr;




typedef struct fontchunkstr
{

 chardatastr    chardata[256-32];

} fontchunkstr;


//require nchars to be incremented by c2-c1 for chardata() to work correctly

static os_error * readchardata(int osfont,int c1,int c2,chardatastr * cd)
{
 os_error * err;
 int        fh;
 int        c;

 CHECKOSFONT(osfont);

// cprintf("readchardata %d %d",c1,c2);

 err=validfont(osfont,0,fontsos[osfont].designsize,0,&fh,NULL);
 if(!err)
 {
  font_setfont2(fh,fonthdcptr);

  for(c=c1;c<c2;c++)
  {
   if(txunicode) font_chardata162(c,&cd->fi,&cd->w,&cd->h,fonthdcptr);
   else          font_chardata2(c,&cd->fi,&cd->w,&cd->h,fonthdcptr);

   cd->c=c;
   cd++;
   fontsos[osfont].nchars++;
  }
 }

 return(err);
}




static int kdcmp(const void * p1,const void * p2)
{
 font_kerningdatastr * kd1;
 font_kerningdatastr * kd2;

 kd1=(font_kerningdatastr *)p1;
 kd2=(font_kerningdatastr *)p2;

 if(kd1->first==kd2->first) return(kd1->second-kd2->second);
 return(kd1->first-kd2->first);

}



os_error * fontreadfontdata(int osfont)
{
 os_error     * err;
 int            fh;
 font_metric    fm;
 fontchunkstr * fp;
 int            npairs;
 int          * glyphs;
 int            nglyph;

 err=NULL;

 CHECKOSFONT(osfont);

 if(!fontsos[osfont].dataread)
 {
  err=validfont(osfont,0,fontsos[osfont].designsize,0,&fh,NULL);
  if(!err)
  {
   font_setfont2(fh,fonthdcptr);
   if(!err) err=font_getmetric2(&fm,fonthdcptr);
   if(!err)
   {
    font_glyphs(fonthdcptr,&glyphs,&nglyph);

#ifdef NEVER    
				if(glyphs)
				{
					int i=0;

					while(1)
					{
						cprintf("start %d range %d",glyphs[i],glyphs[i+1]);
						if(glyphs[i]==0 && glyphs[i+1]==0) break;
						i+=2;
					}
				}
#endif

    fontsos[osfont].ascent=fm.ascent;
    fontsos[osfont].descent=fm.descent;
    fontsos[osfont].height=fm.height;
    fontsos[osfont].intleading=fm.intleading;
    fontsos[osfont].extleading=fm.extleading;
    fontsos[osfont].maxwidth=fm.maxwidth;
    fontsos[osfont].averagewidth=fm.averagewidth;
    fontsos[osfont].dataread=1;
    fontsos[osfont].glyphs=glyphs;
    fontsos[osfont].nglyph=nglyph;

    err=salloc((flex_ptr)&fp,sizeof(fontchunkstr));
    if(!err)
    {
     fontsos[osfont].data=(void*)fp;
     readchardata(osfont,32,256,((fontchunkstr*)fontsos[osfont].data)->chardata);
    }

    if(!err)
    {
     if(txunicode) err=font_kerningdata162(NULL,&npairs,fonthdcptr);
     else          err=font_kerningdata2(NULL,&npairs,fonthdcptr);

     if(!err)
     {
      err=salloc((flex_ptr)&fp,npairs*sizeof(font_kerningdatastr));
      if(!err)
      {
       if(txunicode) err=font_kerningdata162((font_kerningdatastr*)fp,&npairs,fonthdcptr);
       else          err=font_kerningdata2((font_kerningdatastr*)fp,&npairs,fonthdcptr);
      }

      if(!err)
      {
       fontsos[osfont].kdata=(void*)fp;
       fontsos[osfont].npairs=npairs;

       qsort(fp,npairs,sizeof(font_kerningdatastr),kdcmp);
      }
     }
    }
   }
  }
 }


 return(err);
}


static int chcmp(const void * p1,const void * p2)
{
 chardatastr * cd1;
 chardatastr * cd2;

 cd1=(chardatastr*)p1;
 cd2=(chardatastr*)p2;

 return(cd1->c-cd2->c);
}


#define FDCHUNK 32





static os_error * chardata(int font,int c,chardatastr ** cdp)
{
 os_error    * err;
 chardatastr * cd;
 chardatastr   key;
 int           chunk;
 int           n;

 err=NULL;

/*
  if(fontsos[font].charset==SYMBOL_CHARSET)
	{
		c=code16index(c);
		if(c<32) c=32;									// missing glyphs...
	}

*/

// cprintf("char data in c %d",c);

 CHECKOSFONT(font);

 if(c<256)
 {
  *cdp=&((fontchunkstr*)fontsos[font].data)->chardata[c-32];
 }
 else
 {
  key.c=c;
  
  cd=(chardatastr*)bsearch(&key,((fontchunkstr*)fontsos[font].data)->chardata,fontsos[font].nchars,sizeof(chardatastr),chcmp);
  if(!cd)
  {
   err=srealloc((flex_ptr)&fontsos[font].data,(fontsos[font].nchars+FDCHUNK)*sizeof(chardatastr));
   if(!err)
   {
     chunk=(c/FDCHUNK)*FDCHUNK;
    key.c=chunk;
    cd=(chardatastr*)bnearest(&key,((fontchunkstr*)fontsos[font].data)->chardata,fontsos[font].nchars,sizeof(chardatastr),chcmp);
    n=cd-((fontchunkstr*)fontsos[font].data)->chardata;
    memmove(cd+FDCHUNK,cd,sizeof(chardatastr)*(fontsos[font].nchars-n));
    readchardata(font,chunk,chunk+FDCHUNK,cd);
    // at this point readchardata had better have incremented nchars by FDCHUNK
    // worry about what happens to any dud characters?
    cd=cd+(c-chunk);
   }
   else
   {
 //   cprintf("realloc fails %s nchars %d",err->errmess,(fontsos[font].nchars+FDCHUNK)*sizeof(chardatastr));

   }
  }

  *cdp=cd;
 }


// cprintf("chardata c=%d cd %x",c,*cdp);
//	if(*cdp) cprintf("w %d h %d",(*cdp)->w,(*cdp)->h);


 return(err);
}



static os_error * losefontdata(int font)
{
 os_error * err;

 CHECKOSFONT(font);

 err=NULL;

 if(fontsos[font].dataread)
 {
  err=sfree((flex_ptr)&fontsos[font].data);
  if(!err) err=sfree((flex_ptr)&fontsos[font].kdata);
  if(!err) err=sfree((flex_ptr)&fontsos[font].glyphs); 
 }

 return(err);
}




int fontkernshift(int osfont,int first,int second)
{
 font_kerningdatastr * kd;
 font_kerningdatastr * hit;
 font_kerningdatastr   key;
 int                   shift;

 shift=0;
 
 kd=(font_kerningdatastr*)fontsos[osfont].kdata;
 key.first=(unsigned short)first;
 key.second=(unsigned short)second;

// cprintf("kernshift in %d %d",first,second);

 hit=(font_kerningdatastr*)bsearch(&key,kd,fontsos[osfont].npairs,sizeof(font_kerningdatastr),kdcmp);
 if(hit) shift=hit->shift;

// cprintf("kernshift out");

 return(shift);
}



os_error * fontplacement(int osfont,int xpts,int ypts,font_string * fs,int * exts,int xflags,int dx,int ddx)
{
 os_error    * err;
 char        * p;
 char        * lt;
 int           x;
 int           c;
 int           w;
 int           designsize;
 int         * ex;
 int           extent;
	int           nearest;
 int           kern;
 int           lastc;
 int           kdx;
 int           uni;
 chardatastr * cd;


// cprintf("dx=%d ddx=%d",dx,ddx);

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
 {
  designsize=fontsos[osfont].designsize;
  ex=exts;
  extent=(xflags & FP_EXTENT);
		nearest=(xflags & FP_NEAREST);
  kern=(xflags & FP_KERNING);
  uni=(xflags & FPWCS);
  kdx=0;
  lastc=0;

//  cprintf("ddx %d dx %d kern %d kdx %d",ddx,dx,kern,kdx);

  p=fs->s;
  lt=p+fs->term;
  x=0;

  while(p<lt)
  {
   c=*p++;
   if(uni) c+=(*p++)<<8;

   if(c==0) break;

   if(kern && lastc)
   {
    kdx=fontkernshift(osfont,lastc,c);
    kdx=(kdx*xpts)/designsize;
   }

   err=chardata(osfont,c,&cd);
   if(!err)
   {
    w=(cd->w*xpts)/designsize;

//			cprintf("fp w %d xpts %d c %d osfont %d des %d",w,xpts,c,osfont,designsize);



    if(c==SPACE) w+=dx+ddx;
    else         w+=ddx;

    *ex++=w;
    x+=w;

    if(extent)
    {
     if((x+kdx)>fs->x)
     {
      p--;
 					if(uni) p--;
      x-=*(ex-1);
      break;
     }
    }

    if(nearest)
    {
     if((x-w/2+kdx)>fs->x)
     {
      p--;
 					if(uni) p--;
      x-=*(ex-1);
      break;
     }
    }


    if(kern && lastc)
    {
     x+=kdx;
     *(ex-2)+=kdx;
    }
    lastc=c;

//			cprintf("c %c lastc %c kdx %d ex %d",c,lastc,kdx,*(ex-1));
   }

   if(err) break;
  }

  fs->x=x;
  fs->term=p-fs->s;
 }

 return(err);

 USE(ypts);
}



os_error * fontcharwidth(int osfont,int xpts,int ypts,int c,int * w,int * h)
{
 os_error * err;
 chardatastr * cd;

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
 {
  err=chardata(osfont,c,&cd);
  if(!err)
//		cprintf("xpts %d w %d c %d osfont %d des %d",xpts,(cd->w*xpts)/fontsos[osfont].designsize,c,osfont,fontsos[osfont].designsize);
  {
   if(w) *w=(cd->w*xpts)/fontsos[osfont].designsize;
   if(h) *h=(cd->h*ypts)/fontsos[osfont].designsize;
  }
 }

 return(err);
}



os_error * fontcharbbox(int osfont,int xpts,int ypts,int c,font_info * fi)
{
 os_error    * err;
 chardatastr * cd;

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
 {
  err=chardata(osfont,c,&cd);
  if(!err)
  {
   fi->maxx=(cd->fi.maxx*xpts)/fontsos[osfont].designsize;
   fi->minx=(cd->fi.minx*xpts)/fontsos[osfont].designsize;
   fi->maxy=(cd->fi.maxy*ypts)/fontsos[osfont].designsize;
   fi->miny=(cd->fi.miny*ypts)/fontsos[osfont].designsize;
  }
 }

 return(err);
}




/* fi1 contains bounding box of low bit set */
/* fi2 contains bounding box of high+low bit set */


os_error * fontbbox(int osfont,int xpts,int ypts,font_info * fi1,font_info * fi2)
{
 os_error      * err;
 fontosnamestr * fn;
 int             designsize;

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
 {
  fn=&fontsos[osfont];

  designsize=fn->designsize;

  fi1->minx=0;
  fi1->maxx=(fn->maxwidth*xpts)/designsize;
  fi1->miny=-(fn->descent*ypts)/designsize;
  fi1->maxy=((fn->ascent-fn->intleading)*ypts)/designsize;

  fi2->minx=0;
  fi2->maxx=fi1->maxx;
  fi2->miny=fi1->miny;
  fi2->maxy=(fn->ascent*ypts)/designsize;
 }

 return(err);
}




static int cmprange(const void * s1,const void * s2)
{
 int   target;
 int * range;

 target=*((int*)s1);
 range=(int*)s2;

 if(target<*range) return(-1);
 if(target>(*range+*(range+1))) return(1);
 return(0);
}




os_error * fontglyph(int osfont,int c,int * exists)
{
 os_error * err;
 int      * p;
 int        target[2];

 *exists=0;

 CHECKOSFONT(osfont);

 err=fontreadfontdata(osfont);
 if(!err)
 {
  target[0]=c;
  target[1]=1;

  p=(int*)bsearch(target,fontsos[osfont].glyphs,fontsos[osfont].nglyph,2*sizeof(int),cmprange);
  if(p) *exists=1;      
 }

 return(err);
}



// osfont can be <0 if no font assumed
// input is a RISC OS char set number

os_error * fontgetcode8(int c,int osfont,char * p,int * n)
{
 int code;
 int count;


 if(osfont>=0 && fontsos[osfont].charset!=ANSI_CHARSET)  // change for extra charsets
 {
  CHECKOSFONT(osfont);

  code=code8symbol(c);
  *p=(char)code;
  if(n) *n=1;
 }
 else
 {
  code=code8(c);

  if(code>=0)
  {
   *p=(char)code;
   if(n) *n=1;
  }
  else
  {
  	if(osfont>=0 && fontsos[osfont].charset==SYMBOL_CHARSET)
  	{
    if(n) *n=0;
   }
   else
   { 
    code=altcode8(c);
    if(code<0)
    {
     if(n) *n=0;
    }
    else
    {
     count=0;
     while(1)
     {
      *p++=(char)(code & 0xFF);
      code=code>>8;
      count++;
      if(code==0 || count>=4) break;
     }

     if(n) *n=count;
    }
   }
  }
 }

 return(NULL);
}



// input RISC OS 8 bit charset number

os_error * fontgetcode16(int c,int osfont,wchar_t * p,int * n)
{
 int code;
 int exists;
 int count;

 if(osfont>=0 && fontsos[osfont].charset!=ANSI_CHARSET) // change for extra charsets
 {
  CHECKOSFONT(osfont);

  code=code8symbol(c);
  code=code16wincharset(code,fontsos[osfont].charset);
 
  *p=(wchar_t)code;
  if(n) *n=1;
 }
 else
 {
  code=code16(c);

  if(code==UNIBLANK)
  {
   if(n) *n=0;
  }
  else
  {
   if(osfont>=0) fontglyph(osfont,code,&exists);
   else          exists=1;

   if(!exists)
   {
//   cprintf("code %d exists %d",code,exists);

    code=altcode8(c);
    if(code<0)
    {
     if(n) *n=0;
    }
    else
    {
     count=0;
     while(1)
     {
      *p++=(wchar_t)(code & 0xFF);
      code=code>>8;
      count++;
      if(code==0 || count>=4) break;
     }

     if(n) *n=count;
    }
   }
   else
   {
    *p=(wchar_t)code;
    if(n) *n=1;
   }
  }
 }

 return(NULL);
}





/**************************************************************************/





// if in same unit do compare
// if in different units, compare unit names

/*

static int fontoscmp(const void * p1,const void * p2)
{
 fontosnamestr * f1;
 fontosnamestr * f2;
 char        * q1;
 char        * q2;

 f1=(fontosnamestr*)p1;
 f2=(fontosnamestr*)p2;


 if(f1->uoffset==f2->uoffset)
 {
  q1=f1->offset+(char*)fontsos;
  q2=f2->offset+(char*)fontsos;
 }
 else
 {
  q1=f1->uoffset+(char*)fontsos;
  q1+=strlen(q1)+1;
  q2=f2->uoffset+(char*)fontsos;
  q2+=strlen(q2)+1;
 }

 return(cstrcmp(q1,q2));
}

*/


/*
static void sortosfonts(void)
{
 fontosnamestr * fp;
 int             i;
 int             uoffset;

 qsort(fontsos,totosfonts,sizeof(fontosnamestr),fontoscmp);

	// take out any fonts which happen to have both ANSI and SYMBOL charsets

 for(i=0;i<(totosfonts-1);i++)
 {
		if(!cstrcmp((char*)fontsos+fontsos[i].offset,(char*)fontsos+fontsos[i+1].offset))
		{
//   cprintf("font hit %s",(char*)fontsos+fontsos[i].offset);

			if(fontsos[i].charset==ANSI_CHARSET) fontsos[i+1].ignore=1;
			else	  																														fontsos[i].ignore=1;
		}
	}

 fp=fontsos;
 uoffset=-1;

 for(i=0;i<totosfonts;i++)
 {
  if(fp->uoffset!=uoffset)
  {
   fp->unit=1;
   uoffset=fp->uoffset;
  }
  fp++;
 }
}


*/




/* load up the font names */

/* Problems

   A font has to be created with a charset it supports, otherwise
   Win searches for a font with the charset. So we have to support
   charset data. We have to search for Symbol (e.g. Selwyn) as well
   as ANSI fonts.

   Having opened the font, the charset only affects the first 256 chars
   in non-Unicode mode.

   There is a problem, if we use non-ANSI charsets in non-Unicode mode
   as soon as we move to Unicode they will display differently.

   If we search for DEFAULT_CHARSET then for each font we'll get many
   charsets.

   Guess that all charsets for any font are enumerated one after another.
   So easy to compare this face name with the last to stop duplicates.

*/




static int efuoffset;
static int efoffset;


static int efcmp(const void * list,const void * target)
{
 char * facename1;
 char * fontname1;
 char * facename2;
 char * fontname2;
 fontosnamestr * fptr;
 int    cmp;

 fptr=(fontosnamestr*)list;

 facename1=stringheapstring(fontosnames,fptr->fuoffset);
 fontname1=stringheapstring(fontosnames,fptr->foffset);

 facename2=((char**)target)[0];
 fontname2=((char**)target)[1];

 cmp=cstrcmp(facename1,facename2);
 if(!cmp)
 {
  efuoffset=fptr->fuoffset;
  cmp=cstrcmp(fontname1,fontname2);
  if(!cmp) efoffset=fptr->foffset;
 }

 return(cmp);
}



static os_error * eaddfont(char * facename,char * fontname,int italic,int weight,int size,int charset,int pitch,int family,int type,int bold,int regular)
{
 os_error      * err;
 int             i=0;
 char *          targets[2];
 int             result;
 fontosnamestr * fptr;


 err=NULL;

/* searches for nearest match in a sorted list                          */
/* pointer is to entry which key should be inserted before              */
/* if key should be added to end of list, pointer is to end of list + 1 */

 targets[0]=facename;
 targets[1]=fontname;

//  cprintf("eadd %s %s",facename,fontname);

 efoffset=efuoffset=-1;

 fptr=(fontosnamestr*)bnearest2(targets,fontsos,totosfonts,sizeof(fontosnamestr),efcmp,&result);
 i=fptr-fontsos;
 if(result)
 {
  err=flex_chunk((flex_ptr)&fontsos,(totosfonts+1)*sizeof(fontosnamestr),FOCHUNK);
  if(!err)
  {
			fptr=i+fontsos;				// because memory may have moved

   memmove(fptr+1,fptr,sizeof(fontosnamestr)*(totosfonts-i));

   if(efuoffset<0) stringheapadd(&fontosnames,facename,&efuoffset);
   if(efoffset<0)  stringheapadd(&fontosnames,fontname,&efoffset);

   totosfonts++;
  }
 }
 else
 {
  if(charset==ANSI_CHARSET) result=1;
  else
  if(charset==SYMBOL_CHARSET && fontsos[i].charset!=ANSI_CHARSET) result=1;

//  if(result) cprintf("over write charset");
//  else       cprintf("skip charset");

  efuoffset=fontsos[i].fuoffset;
  efoffset=fontsos[i].foffset;
 }

 if(!err)
 {
  if(result)
  {
   fontsos[i].weight=weight;
//      fontsos[i].bold=(fontsos[i].weight==FD_BOLD);
			fontsos[i].bold=bold;
   fontsos[i].italic=italic;
   fontsos[i].fuoffset=efuoffset;
   fontsos[i].foffset=efoffset;
//      fontsos[i].normal=(fontsos[i].weight==FD_REGULAR) && !fontsos[i].italic;
			fontsos[i].normal=regular;
   fontsos[i].ignore=fontignorefont(stringheapstring(fontosnames,efoffset));
   fontsos[i].designsize=size;
   fontsos[i].dataread=0;
   fontsos[i].data=NULL;
   fontsos[i].kdata=NULL;
   fontsos[i].glyphs=NULL;  // DP 16/10/07
   fontsos[i].nchars=0;
   fontsos[i].charset=charset;
   fontsos[i].pitch=pitch;
   fontsos[i].family=family;
   fontsos[i].type=type;
  }
 }

 return(err);
}





static os_error * efont2(char * facename,char * fullname,font_enumdata * data)
{
 os_error * err;

// cprintf("facename %s %d %d em %d ht %d w %d",facename,data->weight,data->italic,data->emsize,data->emheight,data->emwidth);

// cprintf("efont2 face %s font %s charset %d",facename,fullname,data->charset);

 err=eaddfont(facename,fullname,data->italic,data->weight,data->emsize,data->charset,data->pitch,data->family,data->type,data->bold,data->regular);

 return(err);
}



static os_error * efont(char * facename,char * fullname,font_enumdata * data)
{
 os_error         * err;
 font_description   fd;

//	efprint=(strcmp(facename,"Times")==0);


// cprintf("efont face %s font %s charset %d",facename,fullname,data->charset);

// next line to skip FONTENUM_T1
// if(data->type==FONTENUM_TT || data->type==FONTENUM_OT)
 {
  xzstrncpy(fd.name,facename,sizeof(fd.name));
  fd.charset=data->charset;
  err=font_enumerate(&fd,efont2);
 }

 return(err);

 USE(fullname);
}


/*

 HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion\FontSubstitutes

	Problem 28/1/2006

 If we have the font Times installed (HP printers) then the members of Times New Roman
	appear as members of it.

	It seems that enumerating a face name of Times will give the members of TNR
	similarly enumerating Helvetica produces Arial members even if the font is not installed

	Working assumption is that this is controlled by the above reg. key

	Simple case is that a face name only maps to a substitute.

	Horror cases are the substitute mapping back to the original face name.

	Another possibility not related to the above is two fonts with the same face name.


*/



static int ftcmpfn(const void * p1,const void * p2)
{
 return(_tcsicmp(((fontsubtablestr*)p1)->facename,((fontsubtablestr*)p2)->facename));
}


static os_error * tidyosfonts(void)
{
	os_error        * err;
 fontosnamestr   * fp;
 fontosnamestr   * jp;
 fontsubtablestr	* table;
	int               ntable;
	int               i;
	int               j;
	int               k;
	int               l;
 fontsubtablestr			key;
 fontsubtablestr	*	target;
 fontsubtablestr	*	target2;


 err=font_subtable(&table,&ntable);
	if(!err && table && ntable)
	{
  fp=fontsos;
  for(i=0;i<totosfonts;i++)
  {
 		if(fp->unit)
 		{
				key.facename=stringheapstring(fontosnames,fp->fuoffset);

				target=(fontsubtablestr*)bsearch(&key,table,ntable,sizeof(fontsubtablestr),ftcmpfn);
				if(target)
				{
				 key.facename=target->subname;
					target2=(fontsubtablestr*)bsearch(&key,table,ntable,sizeof(fontsubtablestr),ftcmpfn);

//					cprintf("hit %s",key.facename);

					if(!target2)
					{
   			// Simple case
	   		// take out any members of the substitute font in this face

 					// have we seen the substitute face?

      jp=fontsos;
 					for(j=0;j<totosfonts;j++)
 					{
 						if(jp->unit)
 						{
 						 if(!_tcsicmp(stringheapstring(fontosnames,jp->fuoffset),target->subname))
 							{
									//	take out at i any members at j which are at i;

									k=0;
									while(1)
									{
										l=0;
										while(1)
										{
//											cprintf("check %s %s",stringheapstring(fontosnames,(jp+k)->foffset),stringheapstring(fontosnames,(fp+l)->foffset));
											if(!_tcsicmp(stringheapstring(fontosnames,(jp+k)->foffset),stringheapstring(fontosnames,(fp+l)->foffset)))
											{
												(fp+l)->ignore=1;
												break;
											}
											l++;
										 if(l>=totosfonts || (fp+l)->unit) break;
										}
										k++;
										if(k>=totosfonts || (jp+k)->unit) break;
									}

	 						}
	 					}
						jp++;
	 				}
     }
				}
 		}
   fp++;
  }

		font_subtablefree(&table);
	}

	return(err);
}




static os_error * getosfonts(void)
{
 os_error       * err;
 int              i;
 font_description fd;
 int              uoffset;
 fontosnamestr  * fp;


 err=flex_alloc((flex_ptr)&fontsos,FOCHUNK);
 if(!err) err=stringheapinit(&fontosnames,FONAMECHUNK);
 {
  totosfonts=0;

  strcpy(fd.name,"");
  fd.charset=DEFAULT_CHARSET;
  err=font_enumerate(&fd,efont);

  if(!err)
  {
   if(!totosfonts) err=geterror(FOXNON);
   else
   {
    fp=fontsos;
    uoffset=-1;

    for(i=0;i<totosfonts;i++)
    {
//     cprintf("font name %d %s",fp->foffset,stringheapstring(fontosnames,fp->foffset));

     if(fp->fuoffset!=uoffset)
     {
      fp->unit=1;
      uoffset=fp->fuoffset;
     }
     fp++;
    }
   }
  }

		if(!err) err=tidyosfonts();

 }

 return(err);
}



static os_error * loseosfonts(void)
{
 os_error * err;
 int        font;

 err=NULL;

 cprintf("loseosfonts %d",totosfonts);

 for(font=0;font<totosfonts;font++)
 {
//  cprintf("losing font %s",stringheapstring(fontosnames,fontsos[font].foffset));
  err=losefontdata(font);
  if(err) break;
 }

// cprintf("loseosfonts1");

 if(!err)
 {
  totosfonts=0;
  err=flex_free((flex_ptr)&fontsos);
 }

// cprintf("loseosfonts2");

 if(!err)
 {
  err=stringheapfinit(&fontosnames);
 }

// cprintf("loseosfontsX");

 return(err);
}




/* return our font number for a given name */

static int findxosfont(char * name)
{
 int    i;

 for(i=0;i<totosfonts;i++)
 {
  if(!xcstrncmp(name,stringheapstring(fontosnames,fontsos[i].foffset),-1)) break;
 }

 if(i>=totosfonts)
 {
  i=-1;
//  cprintf("miss %s",name);
 }

 return(i);
}








/*****************************************************************************/

/*

		Problem... we seem to have in the main map font names like

				Homerton-Medium
and (at the same time)
				Homerton Medium

different fonts. A search with period->space mapping in xcstrcmp may get the wrong one
solution may be that type 1's should auto generate an id with no '-''s in

		Need to think carefully about this.


*/


// first points to identifier in document
// second points to name of fonts 

// we're using this function to both match Acorn style font names with
// Acorn style font names - i.e. both names have periods in
// and also to match Acorn with Windows, where we need spaces to match periods
// so it is important periods match with periods as well as with spaces.
// note there is an Win-win font name match above which does not use this function


static int fontxcstrncmp(char * first,char * second,int n)
{
 int f;
 int s;

//    cprintf("cmp %s %s",first,second);

 while(n--)
 {
  f=toupper(*first++);
  s=toupper(*second++);

  if(f!=s && f=='.') f=' ';

  if(f!=s || !f) return(f-s);
 }

 return(0);
}




/* return our font number for a given name */

int findxuserfont(char * name)
{
 int    i;

 for(i=0;i<totuserfonts;i++)
 {
		if(fontsuser[i].offset!=fontsuser[i].idoffset)
		{
   if(!fontxcstrncmp(name,(char*)fontsuser+fontsuser[i].idoffset,-1)) break;
		}
 }


 if(i>=totuserfonts)
 {
  for(i=0;i<totuserfonts;i++)
  {
	  if(!fontxcstrncmp(name,(char*)fontsuser+fontsuser[i].offset,-1)) break;
		}

  if(i>=totuserfonts)
  {
   i=-1;
   //  cprintf("miss %s",name);
		}
 }

 return(i);
}



int fontuserstyle(int ifn)
{
 return((fontsuser[ifn].italic<<1)|fontsuser[ifn].bold);
}


int fontosstyle(int osfont)
{
 return((fontsos[osfont].italic<<1)|fontsos[osfont].bold);
}



/* return unique font number from internal font number */

int ufont(int fn)
{
 int c;
 int u;

 u=0;
 c=0;
 while(c<fn)
 {
  c++;
  if(c>=totuserfonts) break; /* trouble we did not find internal number */
  if(fontsuser[c].unit) u++;
 }
 return((c<totuserfonts)?u:-1);
}



/* return internal font number from unique font number */

int ifont(int ufn)
{
 int c,u;

 u=0;
 c=0;
 while(u<ufn)
 {
  c++;
  if(c>=totuserfonts) 
  {
   c=-1;
   break;
  } /* trouble did not find unique number */
  if(fontsuser[c].unit) u++;
 }

 return(c);
}



/* point at the i'th. internal font name */

char * fontidentifieri(int i)
{
 return((char*)fontsuser+fontsuser[i].idoffset);
}




char * fontusername(int userfont)
{
 return((char*)fontsuser+fontsuser[userfont].offset);
}


int fontuserpseudo(int userfont)
{
 return(fontsuser[userfont].pseudo);
}


char * fontosfamilyname(int osfont)
{
 return(stringheapstring(fontosnames,fontsos[osfont].fuoffset));
}


char * fontosname(int osfont)
{
 return(stringheapstring(fontosnames,fontsos[osfont].foffset));
}


char * fontosidentifier(int osfont)
{
 return(stringheapstring(fontosnames,fontsos[osfont].foffset));
}


int fontsubcount(int userfont)
{
 int sn;

 sn=0;
 while(++userfont<totuserfonts)
 {
  if(fontsuser[userfont].unit) break;
  sn++;
 }

 return(sn);
}




static char * subname(int fontn,int sub,int name)
{
 char * p;
 char * q;
 char * lastspc;

 p=(char*)fontsuser+fontsuser[fontn].offset;
// if(name && fonts[fontn].name) p+=strlen(p)+1;
 q=(char*)fontsuser+fontsuser[fontn].uoffset;
 q+=strlen(q)+1;

 if(*p && sub)
 {
  lastspc=p;

  while(*p && *p==*q)
  {
   if(*p==' ') lastspc=p;
   p++;
   q++;
  }

  if(*p && *p!=' ') p=lastspc;

  while(*p==' ') p++;
 }

 return(p);
 USE(name);
}



/* point at the i'th. unique font name */

char * fontlocalname2(int c,int sn,int bits)
{
 static char buff[256];
 char      * p=buff;
 char      * q;
 char        temp[8];

 if(c<0) return(NULL);

 if(sn<0)
 {
  p=(char*)fontsuser+fontsuser[c].uoffset;
  p+=strlen(p)+1;
//  if(fonts[c].name) q+=strlen(q)+1;
  xzstrncpy(buff,p,sizeof(buff));
 }
 else
 {
  while(1)
  {
   if(!sn)
   {
    q=subname(c,1,1);
    xzstrncpy(buff,q,sizeof(buff));
    if(!*q) tsprintf(sizeof(buff),buff,"{FOX02}");

    if(bits & 0x1)
    {
     p=temp;
     *p++=' ';
     *p++='(';
     if(fontsuser[c].normal) *p++='N';
     if(fontsuser[c].bold)   *p++='B';
     if(fontsuser[c].italic) *p++='I';
     *p++=')';
     *p++=0;
     if(p>temp+4) strcat(buff,temp);
    }

    break;
   }

   sn--;
   c++;
   if(c>=totuserfonts) return(NULL);
   if(fontsuser[c].unit) return(NULL);
  }
 }

 return(buff);
}



static char * fontlocalname(int ufn)
{
 return(fontlocalname2(ifont(ufn),-1,0));
}


char * fontlocalnamei(int i)
{
 return(fontlocalname2(i,-1,0));
}




/* return font number from font number and style */
// return user font number

static int fontuserfromstylei(int ifn,int bold,int italic)
{
 int           normal;
 int           i;

 normal=!bold && !italic;

 i=ifn;
 do
 {
//  cprintf("i %d b %d i %d n %d targets b %d i %d n %d weight %x",i,
//           fontsuser[i].bold,fontsuser[i].italic,fontsuser[i].normal,
//            bold,italic,normal,fontsos[i].weight);
  
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic && 
     fontsuser[i].normal==(unsigned int)normal) return(i);

 } while(++i<totuserfonts && !fontsuser[i].unit);


// can't trust normal flag - font may not have weight number exactly same as normal value?
 
 i=ifn;
 do
 {
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic 
     ) return(i);
 } while(++i<totuserfonts && !fontsuser[i].unit);



 if(bold) bold=0;
 normal=!bold && !italic;

 i=ifn;
 do
 {
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic && 
     fontsuser[i].normal==(unsigned int)normal) return(i);

 } while(++i<totuserfonts && !fontsuser[i].unit);


 i=ifn;
 do
 {
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic
     ) return(i);

 } while(++i<totuserfonts && !fontsuser[i].unit);



 if(italic) italic=0;
 normal=!bold && !italic;


 i=ifn;
 do
 {
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic && 
     fontsuser[i].normal==(unsigned int)normal) return(i);

 } while(++i<totuserfonts && !fontsuser[i].unit);


 i=ifn;
 do
 {
  if(fontsuser[i].bold==(unsigned int)bold && fontsuser[i].italic==(unsigned int)italic
     ) return(i);

 } while(++i<totuserfonts && !fontsuser[i].unit);


 return(ifn);
}



int fontuserfromstyle(int ufn,int bold,int italic)
{
 return(fontuserfromstylei(ifont(ufn),bold,italic));
}



int fontuserfromstyle2(int ifn,int bold,int italic)
{
 if(fontsuser[ifn].bold==(unsigned int)bold && fontsuser[ifn].italic==(unsigned int)italic) return(ifn);
 else return(fontuserfromstylei(ifont(ufont(ifn)),bold,italic));
}




int findxuserfontstyle(char * name,int * bold,int * italic)
{
 int ifont;

 ifont=findxuserfont(name);
 if(ifont>=0)
 {
  *bold=fontsuser[ifont].bold;
  *italic=fontsuser[ifont].italic;
 }
 return(ifont);
}



void finduserfontstyle(int ifn,int * bold,int * italic)
{
 *bold=fontsuser[ifn].bold;
 *italic=fontsuser[ifn].italic;
}


int fontosfromuser(int userfont)
{
 if(userfont>=0 && userfont<totuserfonts)
 {
//		cprintf("osfromuser userfont %d result %d",userfont,fontsuser[userfont].osfont);
  return(fontsuser[userfont].osfont);
 }
 return(0);
}





static os_error * loseuserfonts(void)
{
// cprintf("loseuserfonts");


 flex_free((flex_ptr)&fontsuser);

//  cprintf("loseuserfontsX");

 return(NULL);
}






// void fontmap(string & family,string & member,string & identifier,int flags,string & newidentifier,string & spare,int spare1,int spare2);

#define FONTMAPCHUNK 0x200

char * fontmaps;
int    fontmaplen;
int    fontmapactive=1;


int fontmapstate(int fp)
{
 int old;
 old=fontmapactive;
 fontmapactive=stack[fp];
 return(old);
}



// format in memory <family> <member> <identifier> <new identifier> <flags>



void fontmap(int fp)
{
 os_error * err;
 char     * p1;
 char     * p2;
 char     * p3;
 char     * p4;
 int        len;
 int        len1;
 int        len2;
 int        len3;
 int        len4;

 err=NULL;

 p1=stringptr(stack[fp]);          // family must be none null
 len=len1=strlen(p1)+1;
 p2=stringptr(stack[fp+1]);        // font member must be none null
 len2=strlen(p2)+1;
 len+=len2;
 len2+=len1;
 p3=stringptr(stack[fp+2]);        // identifier can be null
 len3=strlen(p3)+1;
 len+=len3;
 len3+=len2;
 p4=stringptr(stack[fp+4]);        // Win id must be none null
 len4=strlen(p4)+1;
 len+=len4;
 len4+=len3;

 len++;  // for flags byte

 if(len1>1 && len2>1 && len4>1)
 {
  err=flex_chunk((flex_ptr)&fontmaps,fontmaplen+len,FONTMAPCHUNK);
  if(!err)
  {
   strcpy(fontmaps+fontmaplen,p1);
   strcpy(fontmaps+fontmaplen+len1,p2);
   strcpy(fontmaps+fontmaplen+len2,p3);
   strcpy(fontmaps+fontmaplen+len3,p4);
   *(fontmaps+fontmaplen+len4)=(char)stack[fp+3];
   fontmaplen+=len;
  }
 }

 zreport(err);
}



void fontremovemap(int fp)
{
 os_error * err;
 char     * member;
 char     * pfam;
 char     * pmem;
 char     * pid;
 char     * pnid;
 char     * pf;
 char     * lt;


 err=NULL;
    
 member=stringptr(stack[fp]);

 // format in memory <family> <member> <identifier> <new identifier> <flags>
   
 if(fontmaps)
 {
  pfam=fontmaps;
  lt=fontmaps+fontmaplen;

//  cprintf("member %s",member);

  while(pfam<lt)
  {
   pmem=pfam+(strlen(pfam)+1);
   pid=pmem+(strlen(pmem)+1);
   pnid=pid+(strlen(pid)+1);
   pf=pnid+(strlen(pnid)+1);

//   cprintf("pmem %s member %s",pmem,member);

   if(cstrcmp(pmem,member)==0)
   {
    err=flex_midextend((flex_ptr)&fontmaps,pf+1-fontmaps,-(pf+1-pfam));
    fontmaplen-=pf+1-pfam;
    break;
   }

   pfam=pf+1;
  }
 }

 zreport(err);
}



void fontclearmap(int fp)
{
 flex_free((flex_ptr)&fontmaps);
 fontmaplen=0;
 USE(fp);
}




// void fontignore(string & identifier,int spare1);
// void fontmap(string & family,string & member,string & identifier,int flags,string & newidentifier,int spare1,int spare2);
// void fontreplace(string & newidentifier,string & oldidentifier,int spare1);


// simple list of strings of fonts to ignore

#define FONTIGNORECHUNK 0x100

char * fontignores;
int    fontignorelen;
int    fontignoreactive=1;


int fontignorestate(int fp)
{
 int old;
 old=fontignoreactive;
 fontignoreactive=stack[fp];
 return(old);
}



// void fontignore(string & identifier);

void fontignore(int fp)
{
 os_error * err;
 char     * p;
 int        len;

 err=NULL;

 p=stringptr(stack[fp]);
 len=strlen(p)+1;
 if(len)
 {
  err=flex_chunk((flex_ptr)&fontignores,fontignorelen+len,FONTIGNORECHUNK);
  if(!err)
  {
   strcpy(fontignores+fontignorelen,p);
   fontignorelen+=len;
  }
 }

 zreport(err);
}




static int fontignorefont(char * fontid)
{
 char * p;
 char * lt;

//	cprintf("font ig %s",fontid);

 if(fontignores && fontignoreactive)
 {
  p=fontignores;
  lt=fontignores+fontignorelen;

  while(p<lt)
  {
//			cprintf("ig %s %s",fontid,p);
   if(cstrcmp(p,fontid)==0)
   {
//				cprintf("ig hit");
    return(1);
   }

   p+=strlen(p)+1;
  }
 }

 return(0);
}




void fontremoveignore(int fp)
{
 os_error * err;
 char     * p;
 char     * p1;
 char     * lt;
 int        len;

 err=NULL;

 p1=stringptr(stack[fp]);
 len=strlen(p1)+1;
 if(len)
 {
  p=fontignores;
  lt=fontignores+fontignorelen;

  while(p<lt)
  {
   len=strlen(p)+1;
   if(cstrcmp(p,p1)==0)
   {
    err=flex_midextend((flex_ptr)&fontignores,(p-fontignores)+len,-len);
    fontignorelen-=len;
    break;
   }
   p+=len;
  }
 }

 zreport(err);
}


void fontclearignore(int fp)
{
 flex_free((flex_ptr)&fontignores);
 fontignorelen=0;
 USE(fp);
}






// font names to postscript, format in memory <old string> <new string>

#define FONTPOSTSCRIPTCHUNK 0x200

char * fontpostscript;
int    fontpostscriptlen;
int    fontpostscriptactive=1;


int fontpostscriptstate(int fp)
{
 int old;
 old=fontpostscriptactive;
 fontpostscriptactive=stack[fp];
 return(old);
}



// void fontremovepostscript(string & oldidentifier);

static os_error * fontremovepostscriptsub(char * p1)
{
 os_error * err;
 char     * p;
 char     * lt;
 int        len;
 int        len1;

 err=NULL;


 if(fontpostscript)
 {
  p=fontpostscript;
  lt=fontpostscript+fontpostscriptlen;
  while(p<lt)
  {
   len=strlen(p)+1;
   if(!cstrcmp(p,p1))
   {
    len1=strlen(p+len)+1;
    {
     err=flex_midextend((flex_ptr)&fontpostscript,(p-fontpostscript)+len+len1,-(len+len1));
     fontpostscriptlen-=len+len1;
     break;
    }
   }
   p+=len;
   p+=strlen(p)+1;
  }
 }

 return(err);
}



void fontremovepostscript(int fp)
{
 os_error * err;
	char     * p1;

 p1=stringptr(stack[fp]);

	err=fontremovepostscriptsub(p1);

 zreport(err);
}




// void fontpostscript(string & newidentifier,string & oldidentifier,int spare);


void fontpostscriptx(int fp)
{
 os_error * err;
 char     * p1;
 char     * p2;
 int        len;
 int        len1;

 err=NULL;

 p1=stringptr(stack[fp+1]);
 len=len1=strlen(p1)+1;
 p2=stringptr(stack[fp]);
 len+=strlen(p2)+1;

	err=fontremovepostscriptsub(p1);
	if(!err)
	{
  if(*p2)
  {
   err=flex_chunk((flex_ptr)&fontpostscript,fontpostscriptlen+len,FONTPOSTSCRIPTCHUNK);
   if(!err)
   {
    strcpy(fontpostscript+fontpostscriptlen,p1);
    strcpy(fontpostscript+fontpostscriptlen+len1,p2);
    fontpostscriptlen+=len;
   }
  }
	}

 zreport(err);
}



char * findpostscriptfont(char * fontid)
{
 char * p;
 char * lt;
 int    len;


 if(fontpostscript && fontpostscriptactive)
 {
  p=fontpostscript;
  lt=fontpostscript+fontpostscriptlen;
  while(p<lt)
  {
   len=strlen(p)+1;

   if(!fontxcstrncmp(fontid,p,-1))
   {
    return(p+len);
   }
   p+=len;
   p+=strlen(p)+1;
  }
 }

 return(NULL);
}






void fontclearpostscript(int fp)
{
 flex_free((flex_ptr)&fontpostscript);
 fontpostscriptlen=0;
 USE(fp);
}













// font names to replace, format in memory <old string> <new string>

#define FONTREPLACECHUNK 0x200

char * fontreplaces;
int    fontreplacelen;
int    fontreplaceactive=1;


int fontreplacestate(int fp)
{
 int old;
 old=fontreplaceactive;
 fontreplaceactive=stack[fp];
 return(old);
}


// void fontreplace(string & newidentifier,string & oldidentifier,int spare);
// oldidentifier can be ""

void fontreplace(int fp)
{
 os_error * err;
 char     * p1;
 char     * p2;
 int        len;
 int        len1;

 err=NULL;

 p1=stringptr(stack[fp+1]);
 len=len1=strlen(p1)+1;
 p2=stringptr(stack[fp]);
 len+=strlen(p2)+1;
 if(*p2)
 {
  err=flex_chunk((flex_ptr)&fontreplaces,fontreplacelen+len,FONTREPLACECHUNK);
  if(!err)
  {
   strcpy(fontreplaces+fontreplacelen,p1);
   strcpy(fontreplaces+fontreplacelen+len1,p2);
   fontreplacelen+=len;
  }
 }

 zreport(err);
}




int findreplacementfont(char * fontid)
{
 char * p;
 char * lt;
 int    defltn;
 int    len;
 int    fontn;

//	cprintf("find replacement font %s",fontid);

 if(fontreplaces && fontreplaceactive)
 {
  defltn=-1;
  p=fontreplaces;
  lt=fontreplaces+fontreplacelen;
  while(p<lt)
  {
   len=strlen(p)+1;

//			cprintf("test %s %s len=%d",fontid,p,len);

   if(len==1)
   {
    if(defltn<0) defltn=findxuserfont(p+1);
   }
   else
   {
//			cprintf("test2 %s %s",fontid,p);

    if(!fontxcstrncmp(fontid,p,-1))
    {

     fontn=findxuserfont(p+len);

//					cprintf("hit %d",fontn);
     
     if(fontn>=0) return(fontn);
    }
   }
   p+=len;
   p+=strlen(p)+1;
  }

  if(defltn>=0) return(defltn);
 }

 return(-1);
}


// void fontremovereplace(string & newidentifier,string & oldidentifier);

void fontremovereplace(int fp)
{
 os_error * err;
 char     * p;
 char     * p1;
 char     * p2;
 char     * lt;
 int        len;
 int        len1;

 err=NULL;

 p1=stringptr(stack[fp+1]);
 p2=stringptr(stack[fp]);

 if(fontreplaces)
 {
  p=fontreplaces;
  lt=fontreplaces+fontreplacelen;
  while(p<lt)
  {
   len=strlen(p)+1;
   if(!cstrcmp(p,p1))
   {
    len1=strlen(p+len)+1;
    if(!cstrcmp(p+len,p2))
    {
     err=flex_midextend((flex_ptr)&fontreplaces,(p-fontreplaces)+len+len1,-(len+len1));
     fontreplacelen-=len+len1;
     break;
    }
   }
   p+=len;
   p+=strlen(p)+1;
  }
 }

 zreport(err);
}



void fontclearreplace(int fp)
{
 flex_free((flex_ptr)&fontreplaces);
 fontreplacelen=0;
 USE(fp);
}





// if in same unit do compare
// if in different units, compare unit names

static int fontusercmp(const void * p1,const void * p2)
{
 fontusernamestr * f1;
 fontusernamestr * f2;
 char            * q1;
 char            * q2;

 f1=(fontusernamestr*)p1;
 f2=(fontusernamestr*)p2;


 if(f1->uoffset==f2->uoffset)
 {
  q1=f1->offset+(char*)fontsuser;
  q2=f2->offset+(char*)fontsuser;
 }
 else
 {
  q1=f1->uoffset+(char*)fontsuser;
  q1+=strlen(q1)+1;
  q2=f2->uoffset+(char*)fontsuser;
  q2+=strlen(q2)+1;
 }

 return(cstrcmp(q1,q2));
}




static void sortuserfonts(void)
{
 fontusernamestr * fp;
 int               i;
 int               uoffset;

 qsort(fontsuser,totuserfonts,sizeof(fontusernamestr),fontusercmp);

 fp=fontsuser;
 uoffset=-1;

 for(i=0;i<totuserfonts;i++)
 {
  if(fp->uoffset!=uoffset)
  {
   fp->unit=1;
   uoffset=fp->uoffset;
  }

  fp++;
 }
}





static int xfinduserfamily(char * family,char * stringptr)
{
 int    i;
 char * p;

 for(i=0;i<totuserfonts;i++)
 {
  if(fontsuser[i].offset==fontsuser[i].uoffset)
  {
   p=stringptr+fontsuser[i].offset;
   p+=strlen(p)+1;
   if(!fontxcstrncmp(family,p,-1)) break;
  }
 }

 if(i>=totuserfonts)
 {
  i=-1;
//  cprintf("miss %s",name);
 }

 return(i);
}






static int xfinduser(char * name,char * stringptr)
{
 int    i;

 for(i=0;i<totuserfonts;i++)
 {
  if(!fontxcstrncmp(name,stringptr+fontsuser[i].offset,-1)) break;
 }

 if(i>=totuserfonts)
 {
  i=-1;
//  cprintf("miss %s",name);
 }

 return(i);
}

/*

Capital = delimiter+Capital
BoldItalic in RISC OS conversions become Bold.Italic
In Windows it would become -of course- Bold Italic

Exception is two capitals in a row. That means 'BT' form Bitstream isn't
split.

*/

#ifdef NEVER

// Problem here is t1's having strange looking names. Allowing name
// to be different to id at the user level is not entirely happy
// because there is a menu of Win font names in the font manager
// maybe that there should be id and name at osfont level
// perhaps menu name should come from t1 font file, perhaps above algorithm
// is not good enough.


static void translatet1(char * t1name,char * buffer)
{
	char * w;
 char * r;
	int    ch;

	r=t1name;
	w=buffer;
	while(1)
	{
		ch=*r;
		if(ch==0) break;

		if(r>t1name && isupper(ch))
  {
   if(isupper(*(r-1)) || *(r-1)=='-')
			{
				*w++=ch;
			}
			else
			{
				*w++=' ';
				*w++=ch;
			}
		}
		else
		if(ch=='-') *w++=' ';
		else								*w++=ch;

		r++;
	}

	*w++=0;
}


#endif



#define FONTUCHUNK (8*sizeof(fontusernamestr))
#define FONTUSTRINGCHUNK (0x2000)

static os_error * getuserfonts(void)
{
 os_error * err;
 int        len;
 int        stringptr;
 int        r;
 int        w;
 int        pendingunit;
 int        uoffset=0;
 char     * strings;
	int        shift;
 char     * p;

	err=NULL;



 totuserfonts=0;

 if(!fontsuser) err=flex_alloc((flex_ptr)&fontsuser,totosfonts*sizeof(fontusernamestr));
	else					 				 err=flex_extend((flex_ptr)&fontsuser,totosfonts*sizeof(fontusernamestr));

 if(!err)
 {
  err=flex_alloc((flex_ptr)&strings,stringheaplength(fontosnames));
  if(!err)
  {
   stringptr=0;
			pendingunit=0;

   for(r=w=0;r<totosfonts;r++)
   {
				if(w>=totosfonts) err=flex_chunk((flex_ptr)&fontsuser,(w+1)*sizeof(fontusernamestr),FONTUCHUNK);
				if(err) break;

    if(fontsos[r].unit) pendingunit=1;

//				if(fontsos[r].ignore) cprintf("font ignore %s",(char*)fontsos+fontsos[r].offset);
 
    if(!fontsos[r].ignore)
    {

#ifdef NEVER					
     if(fontsos[r].type==FONTENUM_T1)
					{
						char t1buffer[256];

      translatet1((char*)fontsos+fontsos[r].offset,t1buffer);
						len=strlen(t1buffer)+1;

      fontsuser[w].offset=stringptr;
				 	err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
				 	if(err) break;

      strcpy(strings+stringptr,t1buffer);
      stringptr+=len;
					}
					else
#endif

					{
      fontsuser[w].offset=stringptr;
      p=stringheapstring(fontosnames,fontsos[r].foffset);
      len=strlen(p)+1;
				 	err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
				 	if(err) break;

      strcpy(strings+stringptr,p);
      stringptr+=len;
					}

     if(pendingunit)
     {
      uoffset=fontsuser[w].offset;
						p=stringheapstring(fontosnames,fontsos[r].fuoffset);
      len=strlen(p)+1;
 					err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
 					if(err) break;

      strcpy(strings+stringptr,p);
      stringptr+=len;
     }

#ifdef NEVER
					if(fontsos[r].type==FONTENUM_T1)
					{
      fontsuser[w].idoffset=stringptr;
      len=strlen((char*)fontsos+fontsos[r].offset)+1;
				 	err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
				 	if(err) break;

      strcpy(strings+stringptr,(char*)fontsos+fontsos[r].offset);
      stringptr+=len;
					}
					else
#endif

					{
      fontsuser[w].idoffset=fontsuser[w].offset;
					}

     fontsuser[w].uoffset=uoffset;
     fontsuser[w].osfont=r;

     fontsuser[w].unit=0;
     fontsuser[w].bold=fontsos[r].bold;
     fontsuser[w].italic=fontsos[r].italic;
     fontsuser[w].normal=fontsos[r].normal;

     pendingunit=0;
     w++;
					totuserfonts++;
    }
   }
  
// now merge in any mapped fonts
// look up font name
// exists? - overwrite
// family name exists?

// format in memory <family> <member> <identifier> <new identifier> <flags>
   
   if(fontmaps && fontmapactive)
   {
    char * pfam;
    char * pmem;
    char * pid;
    char * pnid;
    char * pf;
    int    pfamptr;
    int    osfontptr;
    char * lt;

    pfam=fontmaps;
    lt=fontmaps+fontmaplen;

    while(pfam<lt)
    {
     pmem=pfam+(strlen(pfam)+1);
     pid=pmem+(strlen(pmem)+1);
     pnid=pid+(strlen(pid)+1);
     pf=pnid+(strlen(pnid)+1);

     osfontptr=findxosfont(pnid);
     if(osfontptr>=0)
     {
      w=xfinduser(pmem,strings);
      if(w>=0)
      {


      }
      else
      {
 
       pfamptr=xfinduserfamily(pfam,strings);

//       cprintf("generate font %s pfamptr %d",pmem,pfamptr);
       
       w=totuserfonts;
       err=flex_chunk((flex_ptr)&fontsuser,(w+1)*sizeof(fontusernamestr),FONTUCHUNK);
       if(err) break;
  
       fontsuser[w].offset=stringptr;
       len=strlen(pmem)+1;
 	 				err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
 	 				if(err) break;

       strcpy(strings+stringptr,pmem);
       stringptr+=len;

       if(pfamptr<0)
       {
        uoffset=fontsuser[w].offset;
        len=strlen(pfam)+1;
  	 				err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
  	 				if(err) break;

        strcpy(strings+stringptr,pfam);
        stringptr+=len;
       }
       else
       {
        uoffset=fontsuser[pfamptr].uoffset;

       }

       if(*pid)
       {
        fontsuser[w].idoffset=stringptr;
        len=strlen(pid)+1;
 	 				 err=flex_chunk((flex_ptr)&strings,stringptr+len,FONTUSTRINGCHUNK);
  	 				if(err) break;
        strcpy(strings+stringptr,pid);
        stringptr+=len;
       }
       else
       {
        fontsuser[w].idoffset=fontsuser[w].offset;
       }

       fontsuser[w].uoffset=uoffset;
       fontsuser[w].osfont=osfontptr;

       fontsuser[w].unit=0;
       fontsuser[w].bold=(*pf)&0x1;
       fontsuser[w].italic=((*pf)>>1)&0x1;
       fontsuser[w].normal=(!fontsuser[w].bold) && (!fontsuser[w].italic);
       fontsuser[w].pseudo=1;

       totuserfonts++;
      }
     }

     pfam=pf+1;
    }
   }

// now combine strings with fontsuser

//	 	cprintf("totuser %d stringptr %d",totuserfonts,stringptr);

	 	err=flex_extend((flex_ptr)&fontsuser,totuserfonts*sizeof(fontusernamestr)+stringptr);
	 	if(!err)
	 	{
	 		memcpy(((char*)fontsuser)+totuserfonts*sizeof(fontusernamestr),strings,stringptr);

    shift=totuserfonts*sizeof(fontusernamestr);

    for(r=0;r<totuserfonts;r++)
    {
     fontsuser[r].offset=fontsuser[r].offset+shift;
     fontsuser[r].idoffset=fontsuser[r].idoffset+shift;
     fontsuser[r].uoffset=fontsuser[r].uoffset+shift;
// 	 		cprintf("font %s osfont %d",(char*)fontsuser+fontsuser[r].offset,fontsuser[r].osfont);
    }
		 }

	 	flex_free((flex_ptr)&strings);
  }
 }

	if(!err)
 {
  if(!totuserfonts) err=geterror(FOXNON);
	}

 return(err);


#ifdef NEVER

 totuserfonts=totosfonts;

 len=totuserfonts*sizeof(fontusernamestr);


 flex_alloc((flex_ptr)&fontsuser,len+fontlen);

 memcpy(((char*)fontsuser)+len,((char*)fontsos)+totosfonts*sizeof(fontosnamestr),fontlen);

 shift=totosfonts*(sizeof(fontosnamestr)-sizeof(fontusernamestr));

 for(i=0;i<totosfonts;i++)
 {
  fontsuser[i].offset=fontsos[i].offset-shift;
  fontsuser[i].uoffset=fontsos[i].uoffset-shift;
  fontsuser[i].osfont=i;

  fontsuser[i].unit=fontsos[i].unit;
  fontsuser[i].bold=fontsos[i].bold;
  fontsuser[i].italic=fontsos[i].italic;
  fontsuser[i].normal=fontsos[i].normal;  
 }

 return(NULL);

#endif
}








/*****************************************************************************/


static os_error * losefonts(void)
{
 loseosfonts();
 loseuserfonts();

 return(NULL);
}


static os_error * getfonts(void)
{
 os_error * err;

          err=getosfonts();
// if(!err) sortosfonts();
 if(!err) err=getuserfonts();
 if(!err) sortuserfonts();

 return(err);
}




int getfontname(int fp)
{
 char * p;

 p=fontlocalname(stack[fp]);
 if(p)
 {
  assignstring(stack[fp+1],p);
  return(1);
 }
 else return(0);
}



/* int getfontsubname(int fn,int sn,int bits,string & s) */

int getfontsubname(int fp)
{
 char * p;


 p=fontlocalname2(ifont(stack[fp]),stack[fp+1],stack[fp+2]);
 if(p)
 {
  assignstring(stack[fp+3],p);
  return(1);
 }
 else return(0);
}





static int fontshavechanged;

int fontschanged(int fp)
{
 if(!fontshavechanged) checkforfontschange();

 if(fontshavechanged)
 {
  fontshavechanged=0;
  return(1);
 }
 else
  return(0);

 USE(fp);
}






/* a function that is called by some means, when the available fonts */
/* have changed                                                      */
/* if this goes off whilst printing, trouble will ensue              */

static os_error * changefonts(int eventn,int user,void * data,int data1)
{
 os_error * err;
 filestr  * file;

 hourglasson();

 cprintf("change fonts");

 file=firstfile;
 while(file)
 {
  stylefontchange(0,file);
  file=file->next;
 }

 charsfont(0);


 cprintf("change fontsA");

 fontcachetrash();      // print cached 0

 cprintf("change fontsA1");
 losefonts();
 cprintf("change fontsA2");
 fxtrash();


 cprintf("change fontsB");

      err=getfonts();

 fontshavechanged=1;

  cprintf("change fontsC");

 charsfont(1);

 cprintf("change fontsD");

 file=firstfile;
 while(file)
 {
  stylefontchange(1,file);
  file=file->next;
 }

 cprintf("change fontsE");

 txfontschanged();

 cprintf("change fontsX");

 hourglassoff();

 return(err);

 USE(eventn);
 USE(user);
 USE(data);
 USE(data1);
}



void forcefontschange(int fp)
{
	zreport(forcefontschangelo());
	USE(fp);
}



/*
static os_error * cachefonts(int eventn,int user,
                                         viewstr * view,void * data)
{
// fmapdeclare(view->file);
 return(NULL);

 USE(eventn);
 USE(user);
 USE(data);
}
*/



/* void fontsetup(int fontbits,int spare1,int spare 2,int spare3); */

void fontsetup(int fp)
{
 fontbits=stack[fp];
}




static os_error * unicodefonts(int eventn,int user,void * data,int data1)
{
 return(changefonts(eventn,user,data,data1));
}



static os_error * fontdeviceinit(void)
{
 os_error * err;

 err=graphicscreatehdcptr((hdcstr**)&fonthdcptr);
 graphicsscaleboth2(1,1,1,1,(hdcstr*)fonthdcptr);


 return(err);
}



os_error * fontinit(void)
{
 os_error * err;

 hourglasson();


 fontdeviceinit();
 err=getfonts();
 if(!err) fontcacheinit();
 if(!err) fxinit();

 if(!err) err=addevent(EVENT_FONTSCHANGED,(eventfn)changefonts,0);
// if(!err) err=addevent(EVENT_PRCACHE0,(eventfn)cachefonts,0);
 if(!err) err=addevent(EVENT_TXUNICODE,(eventfn)unicodefonts,0);

 if(!err) err=fxdrawinit();

 hourglassoff();

 return(err);
}















