/*->c.scanplus */

#include <stdafx.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>


#include "os.h"
#include "err.h"
#include "fsx.h"
#include "flex.h"
#include "alloc.h"
#include "fs_dirutils.h"
#include "deb.h"
#include "etc.h"

#include "scanplus.h"



static fsscanhandle scansave;
static char * scandirname=NULL;


void startscan(void)
{
 fs_savescan(&scansave);
 fs_startscan();
}


void endscan(void)
{
 sfree((flex_ptr)&scandirname);
 fs_endscan();
 fs_loadscan(scansave);
}


int nextitem(fxstat * f)
{
 int eof;

 fs_nextitem(scandirname,f,"*.*",&eof);

 return(eof==0);
}



os_error * initscan(char * dirname)
{
 os_error * err;
 int        len;

 if(scandirname) cprintf("!!!!!! overwrite scandir name!!!!!!");

 len=strlen(dirname)+1;
 err=salloc((flex_ptr)&scandirname,len);
 if(!err)
 {
  strcpy(scandirname,dirname);
 }

 startscan();

 return(err);
}








typedef struct sortscanstr
{
 int      dirstringn;
 char   * dirstring;
 fstat2 * dirstat2;
 int      ptr;
} sortscanstr;

#ifdef _DEBUG
static int sortscancount;
#endif

os_error * fs_startsortscan(char * dirname,char * wild,int flags,fsscanhandle * handle)
{
 os_error    * err;
 sortscanstr * sptr;


 err=salloc((flex_ptr)&sptr,sizeof(sortscanstr));
 if(!err)
 {
  err=getdirectory(&sptr->dirstringn,&sptr->dirstring,&sptr->dirstat2,dirname,wild,flags);
  if(!err)
  {
#ifdef _DEBUG
   sortscancount++;
#endif
   sortdirectory(sptr->dirstringn,sptr->dirstring,sptr->dirstat2);
   sptr->ptr=0;
   *handle=(fsscanhandle)sptr;
  }
  else
  {
   sfree((flex_ptr)&sptr);
  }
 }

 return(err);
}




os_error * fs_endsortscan(fsscanhandle * handle)
{
 os_error    * err;
 sortscanstr * sptr;

 sptr=(sortscanstr*)(*handle);

 err=freedirectory(&sptr->dirstring,&sptr->dirstat2);
 if(!err)
 {
  err=sfree((flex_ptr)&sptr);
  if(!err)
  {
#ifdef _DEBUG
   sortscancount--;
#endif
   *handle=NULL;
  }
 }

 return(err);
}


void fs_scanstats(void)
{
#ifdef _DEBUG
 cprintf("scancount %d",sortscancount);
#endif
}


os_error * fs_nextsortitem(fsscanhandle handle,fxstat * f,int * eof)
{
 os_error    * err;
 sortscanstr * sptr;

 err=NULL;

 sptr=(sortscanstr*)handle;
 if(sptr->ptr>=sptr->dirstringn) {*eof=1;/*cprintf("next sort eof");*/}
 else
 {
  f->f=sptr->dirstat2[sptr->ptr].f;
  strcpy(f->name,sptr->dirstring+sptr->dirstat2[sptr->ptr].name);
  sptr->ptr++;
  *eof=0;
//  cprintf("next sort %d",sptr->ptr);
 }

 return(err);
}









os_error * fs_startdoublesortscan(char * dirname0,char * dirname1,char * wild,int flags,fsscanhandle * handle)
{
 os_error    * err;
 sortscanstr * sptr;
 int           stringsize;
 int           i;
 int           j;
 int           dirstringn;
 fstat2      * dirstat2;
 char        * dirstring;
 int           cmp;
 int           len;
 int           ftype0;
 int           ftype1;
 char        * dname0;
 char        * dname1;


 err=salloc((flex_ptr)&sptr,sizeof(sortscanstr));
 if(!err)
 {
  ftype0=ftype1=0;

  fs_exists(dirname0,&ftype0);
  fs_exists(dirname1,&ftype1);

  if(ftype0==0 && ftype1>0)
  {
   dname0=dirname1;
   dname1=dirname0;
   SWOP(ftype0,ftype1);
  }
  else
  {
   dname0=dirname0;
   dname1=dirname1;
  }


  err=getdirectory2(&sptr->dirstringn,&stringsize,&sptr->dirstring,&sptr->dirstat2,dname0,wild,flags);
  if(!err)
  {
   for(i=0;i<sptr->dirstringn;i++)
   {
    sptr->dirstat2[i].flags=(dname0==dirname1);
   }

#ifdef _DEBUG
   sortscancount++;
#endif
   sortdirectory(sptr->dirstringn,sptr->dirstring,sptr->dirstat2);

   if(ftype1==2)
   {
    err=getdirectory2(&dirstringn,NULL,&dirstring,&dirstat2,dname1,wild,flags);   
    if(!err)
    {
     sortdirectory(dirstringn,dirstring,dirstat2);
    
     for(i=j=0;j<dirstringn;)
     {
						if(i>=sptr->dirstringn)
      {
       cmp=1;
						}
						else
						{
       cmp=_stricmp(sptr->dirstat2[i].name+sptr->dirstring,dirstat2[j].name+dirstring);
						}

						cprintf("cmp %d %s %s",cmp,sptr->dirstat2[i].name+sptr->dirstring,dirstat2[j].name+dirstring);

      if(cmp==0)
      {
       sptr->dirstat2[i].f=dirstat2[j].f;
       sptr->dirstat2[i].flags=(dname0==dirname0);
       i++;
       j++;
      }
      else
      if(cmp<0)
      {
       i++;
      }
      else  // insert entry j at i
      {
       err=flex_chunk((flex_ptr)&sptr->dirstat2,(sptr->dirstringn+1)*sizeof(fstat2),64*sizeof(fstat2));
       if(err) break;

							cprintf("insert j @ i %s",dirstat2[j].name+dirstring);

       len=strlen(dirstat2[j].name+dirstring)+1;
       err=flex_chunk((flex_ptr)&sptr->dirstring,stringsize+len,0x800);
       if(err) break;

       memmove(sptr->dirstat2+i+1,sptr->dirstat2+i,(sptr->dirstringn-i)*sizeof(fstat2));
       sptr->dirstat2[i].f=dirstat2[j].f;
       sptr->dirstat2[i].flags=(dname0==dirname0);
       sptr->dirstringn++; 
       sptr->dirstat2[i].name=stringsize;
       strcpy(sptr->dirstring+stringsize,dirstring+dirstat2[j].name);
       stringsize+=len;

       i++;
       j++;
      }
     }

     freedirectory(&dirstring,&dirstat2);
    }
   }

   sptr->ptr=0;
   *handle=(fsscanhandle)sptr;

   if(err)
   {
    freedirectory(&sptr->dirstring,&sptr->dirstat2);
   }
  }

  if(err)
  {
   sfree((flex_ptr)&sptr);
  }
 }

 return(err);
}



os_error * fs_nextdoublesortitem(fsscanhandle handle,fxstat * f,int * eof,int * flags)
{
 os_error    * err;
 sortscanstr * sptr;

 err=NULL;

 sptr=(sortscanstr*)handle;
 if(sptr->ptr>=sptr->dirstringn) {*eof=1;/*cprintf("next sort eof");*/}
 else
 {
  f->f=sptr->dirstat2[sptr->ptr].f;
  *flags=sptr->dirstat2[sptr->ptr].flags;
  strcpy(f->name,sptr->dirstring+sptr->dirstat2[sptr->ptr].name);
  sptr->ptr++;
  *eof=0;
//  cprintf("next sort %d",sptr->ptr);
 }

 return(err);
}



os_error * fs_enddoublesortscan(fsscanhandle * handle)
{
 os_error    * err;
 sortscanstr * sptr;

 sptr=(sortscanstr*)(*handle);

 err=freedirectory(&sptr->dirstring,&sptr->dirstat2);
 if(!err)
 {
  err=sfree((flex_ptr)&sptr);
  if(!err)
  {
#ifdef _DEBUG
   sortscancount--;
#endif
   *handle=NULL;
  }
 }

 return(err);
}


